/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article-db.h"
#include "log.h"
#include "prefs.h" /* for data_dir */
#include "util.h"

//=========================================================================

typedef struct
{
	pan_db db;
	gchar* filename;
	const Server* sdata;
	const Group* gdata;
} AHDB;

//=========================================================================

static gchar*
get_filename (const Server *s,
	      const Group  *g)
{
	g_return_val_if_fail (s!=NULL, NULL);
	g_return_val_if_fail (g!=NULL, NULL);

	return g_strdup_printf ("/%s/%s/%s.db", data_dir, s->name, g->name);
}

article_db
ahdb_ref (const Server *s,
	  const Group  *g)
{
	AHDB* ahdb = g_malloc0 ( sizeof(AHDB) );
	ahdb->filename = get_filename ( s, g );
	ahdb->db = pan_db_ref ( ahdb->filename );
	ahdb->sdata = s;
	ahdb->gdata = g;

	return (article_db)ahdb;
}

gboolean
ahdb_exists (const Server *s,
	     const Group  *g)
{
	gchar* filename = get_filename (s, g);
	gboolean exists = file_exists (filename);
	g_free (filename);

	return exists;
}

void
ahdb_delete_db_file (const Server *s,
		     const Group  *g)
{
	gchar* filename = get_filename (s, g);
	remove (filename);
	g_free (filename);
}

void
ahdb_unref (article_db db)
{
	AHDB* ahdb = (AHDB*)db;

	g_free (ahdb->filename);
	ahdb->filename = 0;
	pan_db_unref (ahdb->db);
	ahdb->db = 0;

	g_free (ahdb);
}

//=========================================================================

const char *
ahdb_field (const char *flat_adata,
	    int         offset)
{
	int i = 0;
	const char *p = flat_adata;

	if (offset == 0)
		return p;
	while (*p != '\0') {
		if (*p == '\n')
			i++;
		if (i == offset)
			return (p+1);
		p++;
	}
	return NULL;
}

void
ahdb_save (article_db          db,
	   const Article      *adata,
	   gboolean            sync)
{
	AHDB* ahdb = (AHDB*)db;

	gchar* pch = g_strdup_printf (
		"%d\n" "%lu\n"
		"%d\n" "%d\n"
		"%d\n" "%s\n"

		"%s\n" "%s\n"
		"%s\n"
		"%s\n"
		"%s\n"
		"%s\n",

		(int) adata->date, (unsigned long)(adata->state & ~STATE_NEW),
		adata->linecount, adata->parts,
		adata->part, adata->number,

		adata->subject, adata->author,
		(adata->references ? adata->references : ""),
		(adata->newsgroups ? adata->newsgroups : ""),
		(adata->reply_to ? adata->reply_to : ""),
		(adata->followup_to ? adata->followup_to : "") );

	pan_db_put_value_str (ahdb->db, adata->message_id, pch);

	g_free (pch);

	if (sync)
		ahdb_sync (db);
}

void
ahdb_save_all (article_db     db,
	       Article      **articles,
	       int            article_qty,
	       GFunc          gfunc,
	       gpointer       user_data)
{
	int i;
	for (i=0; i<article_qty; ++i)
	{
		if (gfunc)
			(*gfunc)(articles[i],user_data);

		ahdb_save (db, articles[i], FALSE);
	}

	ahdb_sync (db);
}

void
ahdb_sync (article_db db)
{
	AHDB* ahdb = (AHDB*)db;
	pan_db_sync ( ahdb->db );
}

//=========================================================================

static Article*
article_deserialize (const Group      *gdata,
		     const char       *message_id,
		     const char       *db_text)
{
	gchar** sd;
	Article* adata;
        int i;

	if ( !gdata )
		return NULL;

	/* break up the db_text */
	g_return_val_if_fail (message_id!=NULL, NULL);
	g_return_val_if_fail (db_text!=NULL, NULL);
	sd = g_strsplit ( db_text, "\n", 12 );
	g_return_val_if_fail (sd!=NULL, NULL);

	/* build the article data */
	adata = article_new ();
	adata->message_id = g_strdup (message_id);
	g_return_val_if_fail (sd[0]!=NULL, NULL);

	/* populate the article data */
	adata->date = atoi (sd[0]);
	adata->state = (unsigned long) strtol (sd[1], NULL, 10);
	adata->linecount = atoi (sd[2]);
	adata->parts = atoi (sd[3]);
	adata->part = atoi (sd[4]);
	adata->number = sd[5];
	adata->subject = sd[6];
	adata->author = sd[7];
	adata->references = sd[8];
	adata->newsgroups = sd[9];
	adata->reply_to = sd[10];
	adata->followup_to = sd[11];
	adata->threads = NULL;

	/* cleanup */
        g_free (sd[0]);
        g_free (sd[1]);
        g_free (sd[2]);
        g_free (sd[3]);
        g_free (sd[4]);
        for (i=12; sd[i]!=NULL; ++i) g_free (sd[i]);
        g_free (sd);

	return adata;
}

Article*
ahdb_get (article_db  db,
	  const char *message_id )
{
	AHDB* ahdb = (AHDB*)db;
	gchar* s = pan_db_get_value_str ( ahdb->db, message_id );
	Article* a = article_deserialize ( ahdb->gdata, message_id, s );
	g_free ( s );
	return a;
}

typedef struct
{
	AHDB* ahdb;
	GPtrArray* articles;
	GPtrArray* delete_keys;
	GFunc gfunc;
	gpointer user_data;
	unsigned int first;
        int expire_total_qty;
	int expire_read_qty;
}
GetArticlesStruct;

static void
ahdb_get_all_dbfunc (const char *key,
		     const char *val,
		     gpointer    user_data)
{
	GetArticlesStruct* gas = (GetArticlesStruct*) user_data;
	Article* a = article_deserialize (gas->ahdb->gdata, key, val);

	if (a != NULL)
	{
		/* Expire if it's not on server anymore */
		if (atoi(a->number) < gas->first)
		{
			gas->expire_total_qty++;
			if (article_flag_on (a, STATE_READ))
				gas->expire_read_qty++;

			/* steal a's message_id string */
			g_ptr_array_add (gas->delete_keys, a->message_id);
			a->message_id = NULL;
			pan_object_unref (PAN_OBJECT(a));
		}
		else
		{
			g_ptr_array_add (gas->articles, a);
			if (gas->gfunc)
				(*gas->gfunc)(a, gas->user_data);
		}
	}
}

GPtrArray*
ahdb_get_all (article_db db,
	      GFunc      gfunc,
	      gpointer   user_data)
{
	AHDB* ahdb = (AHDB*)db;
	GetArticlesStruct gas;
	GPtrArray * articles;
	int i;

	/* reserve pointers for the array */
	articles = g_ptr_array_new ();
	pan_g_ptr_array_reserve (articles, group_get_attrib_i (ahdb->sdata, ahdb->gdata, "Total"));

	/* load the articles... */
	gas.articles = articles;
	gas.ahdb = ahdb;
	gas.gfunc = gfunc;
	gas.user_data = user_data;
	gas.expire_read_qty = 0;
	gas.expire_total_qty = 0;
	gas.delete_keys = g_ptr_array_new ();
	gas.first = group_get_attrib_i (ahdb->sdata, ahdb->gdata, "First");
	pan_db_foreach (ahdb->db, ahdb_get_all_dbfunc, (gpointer)&gas);

	if (ahdb->gdata->type == USER_FOLDER)
		return gas.articles;

	/* delete expired articles.  pan_db_foreach locks the db,
	   so we have to do it here instead of in ahdb_get_all_dbfunc. */
	for (i=0; i<gas.delete_keys->len; ++i) {
		char * key = (char*) g_ptr_array_index (gas.delete_keys, i);
		pan_db_erase (gas.ahdb->db, key);
		g_free (key);
	}
	g_ptr_array_free (gas.delete_keys, TRUE);

	group_add_attrib_i (ahdb->sdata, ahdb->gdata, "Read", -gas.expire_read_qty);
	group_add_attrib_i (ahdb->sdata, ahdb->gdata, "Total", -gas.expire_total_qty);

	log_add_va (_("Expired %d articles (%d were marked read) in group %s"),
		gas.expire_total_qty,
		gas.expire_read_qty,
		ahdb->gdata->name );

	return gas.articles;
}

//=========================================================================

static void
ahdb_length_dbfunc (const char *key,
		    const char *val,
		    gpointer    user_data)
{
	++(*((int*)user_data));
}

int
ahdb_length (article_db db)
{
	int i = 0;
	pan_db_foreach (((AHDB*)db)->db, ahdb_length_dbfunc, (gpointer)&i);
	return i;
}

//=========================================================================


void
ahdb_erase (article_db          db,
	    const Article *adata,
	    const char         *message_id)
{
	AHDB* ahdb = (AHDB*)db;
	if (adata)
		pan_db_erase (ahdb->db, adata->message_id);
	else if (message_id)
		pan_db_erase (ahdb->db, message_id);
}

void
ahdb_erase_all (article_db db)
{
	AHDB* ahdb = (AHDB*)db;
	pan_db_erase_all ( ahdb->db );
}

