/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <string.h>

#include <glib.h>
#include <gnome.h>

#include "acache.h"
#include "debug.h"
#include "server.h"
#include "folderlist.h"
#include "group.h"
#include "gui.h"
#include "article.h"
#include "articlelist.h"
#include "util.h"

/***
****
***/

PanCallback *article_deleted = NULL;

/***
****
***/

static void
article_update_impl (article_db db,
		     const Article* adata,
		     gboolean sync,
		     gboolean update_ui);

static void
article_add_flag_impl (article_db db,
		       Article *adata,
		       int flag,
		       gboolean sync,
		       gboolean update_ui);

static void
article_remove_flag_impl (const article_db db,
			  Article* adata,
			  int flag,
			  gboolean sync,
			  gboolean update_ui);

/***
****
***/


static GMemChunk * article_chunk = NULL;

void
article_init (void)
{
	article_deleted = pan_callback_new ();

	article_chunk = g_mem_chunk_create (Article, 1000, G_ALLOC_AND_FREE);
}

void
article_isolate (Article* adata)
{
	GSList *l = NULL;
	Article *parent = adata->parent;

	if (parent)
		parent->threads = g_slist_remove (parent->threads, adata);

	for (l=adata->threads; l!=NULL; l=l->next) {
		Article *child = ARTICLE(l->data);
		child->parent = parent;
		if (parent)
			parent->threads = g_slist_append (parent->threads, child);
	}

	g_slist_free (adata->threads);
	adata->threads = NULL;
	adata->parent = NULL;
}

/**
***
**/

void
article_destructor (PanObject* o)
{
	Article* adata = ARTICLE(o);
        g_return_if_fail (adata!=NULL);

	/* clean up article parts */
	debug (DEBUG_PAN_OBJECT, "article_destructor: %p", adata);
	g_free (adata->number);
	g_free (adata->subject);
	g_free (adata->author);
	g_free (adata->message_id);
	g_free (adata->references);
	g_free (adata->newsgroups);
	g_free (adata->reply_to);
	g_free (adata->followup_to);
	g_slist_free (adata->threads);

	/* clean up parent */
	pan_object_destructor (o);
}

void
article_constructor (Article* adata)
{
	g_return_if_fail (adata!=NULL);

	/* init the parent class */
	pan_object_constructor (PAN_OBJECT(adata), article_destructor);

	/* init the article bits */
	debug (DEBUG_PAN_OBJECT, "article_constructor: %p", adata);
	adata->number = NULL;
	adata->subject = NULL;
	adata->author = NULL;
	adata->message_id = NULL;
	adata->references = NULL;
	adata->date = (time_t)0;
	adata->linecount = 0;
	adata->reply_to = NULL;
	adata->followup_to = NULL;
	adata->newsgroups = NULL;
	adata->parts = 0;
	adata->part = 0;
	adata->parent = NULL;
	adata->threads = NULL;
	adata->unread_children = 0;
	adata->self_passes_filter = FALSE;
	adata->tree_passes_filter = FALSE;
}

Article*
article_new (void)
{
	Article* adata = g_new (Article, 1);
	article_constructor (adata);
	return adata;
}

/***
****
***/

/**
 * @pararm article_db
 * @param adata article to be marked as read
 * @param marked pointer to an int to be incremented for each article marked read
 *
 * If an article is a multipart article, then the multipart articles are also
 * marked as read if they're not read already.
 */
static void
article_mark_read_impl (article_db db,
			Article* adata,
			int* marked)
{
	if (!adata)
		return;

	if (!article_flag_on (adata, STATE_READ)) {
		article_add_flag_impl (db, adata, STATE_READ, FALSE, FALSE);
		articlelist_adata_read_changed (adata);
		++*marked;
	}

	/* special case: mark ALL of a multipart read */
	if (adata->part>0 && adata->parts>1 && adata->threads!=NULL) {
		GSList *l;
		for (l=adata->threads; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_mark_read_impl (db, child, marked);
		}
	}
}


/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks an article as read as described in article_mark_read_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated.
 */
void
article_mark_read (const Server* sdata,
		   const Group* gdata,
		   Article* adata,
		   gboolean sync )
{
	int marked = 0;
	article_db db = NULL;
	
	db = ahdb_ref (sdata, gdata);
	article_mark_read_impl (db, adata, &marked);

	if (sync)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (sdata, gdata, "Read", marked);

	ahdb_unref (db);

}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks a list of articles as read as described in article_mark_read_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
articles_mark_read (
	const Server* sdata,
	const Group* gdata,
	Article** articles,
	int article_qty,
	gboolean sync )
{
	int marked = 0;
	article_db db = NULL;
	int i;

	g_return_if_fail (articles!=NULL);
	g_return_if_fail (article_qty>=1);

	db = ahdb_ref (sdata, gdata);
	for (i=0; i!=article_qty; ++i)
		article_mark_read_impl (db, articles[i], &marked);

	if (sync && marked)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (sdata, gdata, "Read", marked);

	ahdb_unref (db);
}


/**
***
**/

/**
 * @pararm article_db
 * @param adata article to be marked as read
 * @param marked pointer to an int to be incremented for each article marked unread
 *
 * If an article is a multipart article, then the multipart articles are also
 * marked as unread if they're not unread already.
 */
static void
article_mark_unread_impl (article_db db,
			  Article* adata,
			  int *marked )
{
	if (!adata)
		return;

	if (article_flag_on (adata, STATE_READ)) {
		article_remove_flag_impl (db, adata, STATE_READ, FALSE, FALSE);
		articlelist_adata_read_changed (adata);
		++*marked;
	}

	/* special case: mark ALL of a multipart unread */
	if (adata->part>0 && adata->parts>1 && adata->threads!=NULL) {
		GSList *l;
		for (l=adata->threads; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_mark_unread_impl (db, child, marked);
		}
	}
}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks an article as unread as described in article_mark_unread_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
article_mark_unread (const Server* sdata,
		     const Group* gdata,
		     Article* adata,
		     gboolean sync)
{
	int marked = 0;
	article_db db = 0;

	db = ahdb_ref (sdata, gdata);
	article_mark_unread_impl (db, adata, &marked);

	if (sync && marked)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (sdata, gdata, "Read", -marked);

	ahdb_unref (db);
}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * Marks a list of articles as unread as described in article_mark_unread_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
articles_mark_unread (const Server* sdata,
                      const Group* gdata,
                      Article** articles,
                      int article_qty,
                      gboolean sync )
{
	int marked = 0;
	article_db db = 0;
	int i;

	g_return_if_fail (articles!=NULL);
	g_return_if_fail (article_qty>=1);

	db = ahdb_ref (sdata, gdata);
	for (i=0; i!=article_qty; ++i)
		article_mark_unread_impl (db, articles[i], &marked);

	if (sync) {
		ahdb_sync (db);
		if (marked)
			group_add_attrib_i (sdata, gdata, "Read", -marked);
	}

	ahdb_unref (db);
}

/***
****
***/

static void
article_delete_impl (article_db db,
                     Article* adata,
                     int* read,
                     int* total)
{
	if (!adata)
		return;

	/* special case: if a multipart, delete ALL of it */
	if (adata->part>0 && adata->parts>1 && adata->threads!=NULL) {
		GSList *l;
		GSList *tmp = g_slist_copy (adata->threads);
		for (l=tmp; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_delete_impl (db, child, read, total);
		}
		g_slist_free (tmp);
	}

	/* notify listeners before the delete takes place */
	pan_callback_call (article_deleted, adata, NULL);

	/* update counts */
	++*total;
	if (article_flag_on (adata, STATE_READ))
		++*read;

	/* isolate */
	article_isolate (adata);

	/* erase from cache */
	acache_delete (adata->message_id);

	/* erase from db */
	ahdb_erase (db, adata, NULL);

	/* erase from memory */
	pan_object_unref (PAN_OBJECT(adata));
}

void
article_delete (const Server* sdata,
                const Group* gdata,
                Article* adata,
                gboolean sync)
{
	int total = 0;
	int read = 0;
	article_db db = 0;

	/* sanity checks */
	g_return_if_fail (sdata!=NULL);
	g_return_if_fail (gdata!=NULL);
	g_return_if_fail (adata!=NULL);

	db = ahdb_ref (sdata, gdata);
	article_delete_impl (db, adata, &read, &total);
	pan_warn_if_fail (total >= read);

	if (sync)
		ahdb_sync (db);

	group_add_attrib_i (sdata, gdata, "Total", -total);
	group_add_attrib_i (sdata, gdata, "Read", -read);

	ahdb_unref (db);
}

void
articles_delete (const Server* sdata,
                 const Group* gdata,
                 Article** articles,
                 int article_qty,
                 gboolean sync)
{
	int total = 0;
	int read = 0;
	int i;
	article_db db = 0;

	/* sanity checks */
	g_return_if_fail (sdata!=NULL);
	g_return_if_fail (gdata!=NULL);
	g_return_if_fail (article_qty>=0);
	g_return_if_fail (article_qty ? articles!=NULL : 1);

	db = ahdb_ref (sdata, gdata);
	for (i=0; i!=article_qty; ++i)
		article_delete_impl (db, articles[i], &read, &total);

	if (sync)
		ahdb_sync (db);
	if (total)
		group_add_attrib_i (sdata, gdata, "Total", -total);
	if (read)
		group_add_attrib_i (sdata, gdata, "Read", -read);

	ahdb_unref (db);
}

/***
****
***/

static void
article_add_flag_impl (article_db db,
                       Article *adata,
                       int flag,
                       gboolean sync,
                       gboolean update_ui)
{
	g_return_if_fail (adata!=NULL);

	adata->state |= flag;
	article_update_impl (db, adata, sync, update_ui);
}
void
article_add_flag (const Server *sdata,
                  const Group *gdata,
                  Article* adata,
                  int flag,
                  gboolean sync)
{
	article_db db = ahdb_ref (sdata, gdata);
	article_add_flag_impl (db, adata, flag, sync, TRUE);
	ahdb_unref (db);
}


/***
****
***/

static void
article_remove_flag_impl (const article_db db,
                          Article* adata,
                          int flag,
                          gboolean sync,
                          gboolean update_ui)
{
	g_return_if_fail (adata!=NULL);

	adata->state &= ~flag;
	article_update_impl (db, adata, sync, update_ui);
}

void
article_remove_flag (const Server *sdata,
                     const Group *gdata,
                     Article* adata,
                     int flag,
                     gboolean sync)
{
	article_db db = ahdb_ref (sdata, gdata);
	article_remove_flag_impl (db, adata, flag, sync, TRUE);
	ahdb_unref (db);
}


/***
****
***/

static void
article_update_impl (article_db db,
		     const Article* adata,
		     gboolean sync,
		     gboolean update_ui)
{
	/* entry assertions */
	g_return_if_fail (adata && adata->message_id);

	/* update article header db... */
	ahdb_save (db, adata, sync);

	/* update the article ui... */
	if (update_ui)
		articlelist_update_node (adata);
}

void
article_update (const Server *sdata,
		const Group *gdata,
		const Article* adata,
		gboolean sync)
{
	article_db db = ahdb_ref (sdata, gdata);
	article_update_impl (db, adata, sync, TRUE);
	ahdb_unref (db);
}

/**
***
**/

gchar*
article_get_thread_message_id (const Article* adata)
{
	gchar *pch = g_strdup (adata->references);
	gchar *pch2 = strchr (pch, '>');

	if (pch2)
		pch2[1] = '\0';
	else {
		g_free (pch);
		pch = g_strdup (adata->message_id);
	}

	return pch;
}

static void
article_get_entire_thread_count (Article* top, int* qty)
{
	GSList *l;
	++*qty;
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_count (ARTICLE(l->data), qty);
}
static void
article_get_entire_thread_impl (Article* top, Article** a, int *idx)
{
	GSList *l;
	a[*idx] = top;
	++*idx;
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_impl (ARTICLE(l->data), a, idx);
}

void
article_get_entire_thread (Article   *adata,
			   int            *setme_qty,
			   Article ***setme_array)
{
	Article* top = adata;

	*setme_array = NULL;
	*setme_qty = 0;

	if (top != NULL)
	{
		int qty = 0;
		int index = 0;
		Article ** buf = NULL;

		while (top->parent != NULL)
			top = top->parent;

		article_get_entire_thread_count (top, &qty);
		buf = g_new (Article*, qty);
		article_get_entire_thread_impl (top, buf, &index);
		pan_warn_if_fail (index == qty);

		*setme_array = buf;
		*setme_qty = qty;
	}
}

#if 0
/*
 * 1. open source db
 * 2. copy article from source db to folder db
 * 3. erase article from source db
 *
 */
static void
article_move_to_folder (Article *adata,
			Server  *sdata,
			Group   *group,
			Group   *folder)
{
	g_return_if_fail (sdata != NULL);
	g_return_if_fail (adata != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

	article_copy_to_folder (adata, group, folder);

	article_delete (sdata, group, adata, FALSE);
}

void
article_copy_to_folder (Article *adata,
			Group   *group,
			Group   *folder)
{
//	article_db srcdb  = NULL;
	article_db destdb = NULL;

//	g_return_if_fail (sdata != NULL);
	g_return_if_fail (adata != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

//	srcdb = ahdb_ref (sdata, from);
	destdb = ahdb_ref (folder_sdata, folder);

	ahdb_save (destdb, adata, TRUE);

//	ahdb_unref (srcdb);
//	ahdb_sync (destdb);
	ahdb_unref (destdb);
}


typedef struct
{
	Server *sdata;
	Group  *gdata;
	GSList *list_of_adatas;
} ArticleMoveStruct;

void
article_copy_list_to_folder (GSList *articles,
			     Group *group)
{

}

static void
article_move_list_destroy_cb (GtkObject *object, gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	GSList *l;

	for (l=ams->list_of_adatas; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		g_slist_foreach (l2, article_free_gfunc, NULL);
		g_slist_free (l2);
	}
	g_slist_free (ams->list_of_adatas);
	g_free (ams);
}

static void
article_move_list_clicked_cb (GnomeDialog *dialog,
                              int button,
                               gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	Server *sdata = ams->sdata;
	Group  *gdata = ams->gdata;
	Group *folder = NULL;
	GSList *l;

	g_return_if_fail (sdata != NULL);
	g_return_if_fail (gdata != NULL);


	/* Cancel button */
	if (button == 1)
	{
		pan_lock ();
		article_move_list_destroy_cb (GTK_OBJECT (dialog), NULL);
		gtk_widget_destroy (GTK_WIDGET (dialog));
		pan_unlock ();
		return;
	}

	folder = gtk_object_get_data (GTK_OBJECT (dialog), "selectedfolder");

	for (l=ams->list_of_adatas; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		GSList *tmp = NULL;

		for (tmp=l2; tmp!=NULL; tmp=tmp->next)
		{
			Article *adata = ARTICLE(tmp->data);

			article_move_to_folder (adata, sdata, gdata, folder);
		}
	}

	pan_lock ();
	gtk_widget_destroy (GTK_WIDGET (dialog));
	pan_unlock ();

}

void
article_move_list_to_folder (GSList *articles,
                             Server *sdata,
                             Group *group)
{
	GnomeDialog *dialog = folderlist_picker_new ();

	ArticleMoveStruct *ams = g_new (ArticleMoveStruct, 1);

	ams->sdata = sdata;
	ams->gdata = group;
	ams->list_of_adatas = articles;

	/* connect signals */
	gtk_signal_connect (GTK_OBJECT (dialog), "clicked",
			    GTK_SIGNAL_FUNC (article_move_list_clicked_cb),
			    ams);
//	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
//			    GTK_SIGNAL_FUNC (article_move_list_destroy_cb),
//			    ams);
			    
	gui_popup_draw (GTK_WIDGET (dialog), Pan.window);
}
#endif
