/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-exec.h>
#include <libgnome/gnome-mime.h>
#include <libgnome/gnome-mime-info.h>
#include <libgnome/gnome-i18n.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <pthread.h>

#include "../uulib/uudeview.h"
#include "../uulib/fptools.h"

#include "acache.h"
#include "debug.h"
#include "decode.h"
#include "log.h"
#include "prefs.h" /* for download_dir */
#include "util.h"
//#include "status-item.h"
//#include "task-decode.h"

static pthread_mutex_t uulib_lock = PTHREAD_MUTEX_INITIALIZER;

/***
 *** Private Function Prototypes
 ***/

static char* UUD_file_name_filter (void *opaque, char *fname);
static int   UUD_busy_callback (void *opaque, uuprogress *progress);
static void  UUD_message_callback (void *opaque, char *msg, int level);
static void  decode_open (const char *fname);

/***
 *** Public Routines
 ***/


void
decode_article (decode_data *dd)
{
	uulist *item;
	GSList *p;
	gint i = 0;
	gint part_qty = 0;
	gint success = 0;
	gchar *filename = NULL;
	Article* adata = ARTICLE(dd->articles->data);

	/* sanity checking */
	g_return_if_fail (dd!=NULL);
	g_return_if_fail (dd->sdata!=NULL);
	g_return_if_fail (dd->articles!=NULL);

	status_item_emit_progress (dd->item, 0);
	status_item_emit_status (
		dd->item, _("Decoding: %s"), adata->subject);

	/* debug... */
	part_qty = g_slist_length(dd->articles);
	debug(DEBUG_DECODE,
		"starting to decode %d-part article '%s'",
		part_qty,
		adata->subject);

	/**
	***  INIT THE DECODE LIBRARY
	**/

	status_item_emit_status (
		dd->item, _("Waiting to Decode %s"),
		adata->subject );
	pthread_mutex_lock (&uulib_lock);

	/* decide where to save the file */
	directory_check (dd->path);
	chdir (dd->path);

	status_item_emit_status (
		dd->item, _("Preparing %s"), adata->subject);
	UUInitialize ();

	UUSetOption (UUOPT_FAST, part_qty==1 && (int)fast_decode_enabled, NULL);
	UUSetMsgCallback (adata->subject, UUD_message_callback);
	UUSetBusyCallback (dd->item, UUD_busy_callback, 1);
	UUSetFNameFilter (NULL, UUD_file_name_filter);

	/**
	***  LOAD THE ARTICLE(S)
	**/

	for (p=dd->articles, i=0; p!=NULL; p=p->next, ++i)
	{
		const Article* article = ARTICLE(p->data);
		int retval;
		gchar* path = g_strdup_printf (
			"/%s/%s.body", data_dir, article->message_id);
		debug (DEBUG_DECODE,
			"Decoding part %d of %d: %s...", i+1, part_qty, path);
		retval = UULoadFile (path, NULL, 0);
		g_free (path);
		if (retval != UURET_OK)
		{
			/* cleanup */
			UUCleanUp();
			pthread_mutex_unlock (&uulib_lock);
			debug(DEBUG_DECODE,
				"Decode couldn't load tmp file '%s'", path);
			return;
		}
	}
	
	status_item_emit_status (
		dd->item, _("Decoding: %s"), adata->subject);

	/**
	***  DECODE
	**/

	if ((item = UUGetFileListItem(0)) == NULL)
	{
		log_add_va (_("Error 1 decoding '%s'"), adata->subject);
		debug(DEBUG_DECODE,"Error 1 decoding '%s'", adata->subject);
	}
	else if (!(item->state & UUFILE_OK))
	{
		log_add_va (_("Error 2 decoding '%s'"), adata->subject);
		debug(DEBUG_DECODE,"Error 2 decoding '%s'", adata->subject);
	}
	else if ((UUDecodeFile (item, NULL)) != UURET_OK)
	{
		log_add_va (_("Error 3 decoding '%s'"), adata->subject);
		debug(DEBUG_DECODE,"Error 3 decoding '%s'", adata->subject);
	}
	else
	{
		filename = g_strdup(UUFNameFilter(item->filename));
		log_add_va (
			"Decoded '%s' from group '%s', article '%s'",
			filename,
			dd->gdata?dd->gdata->name:"Unknown",
			adata->subject);
		debug (DEBUG_DECODE,
			"Decoded '%s' from group '%s', article '%s'",
			filename,
			dd->gdata?dd->gdata->name:"Unknown",
			adata->subject);
		++success;
		if (dd->filename != NULL)
			rename (item->filename, dd->filename);
	}

	/**
	***  CLEANUP
	**/

	UUCleanUp ();
	pthread_mutex_unlock (&uulib_lock);
	if (dd->open && filename)
		decode_open (filename);
	g_free (filename);

	/* update the node */
	adata->state &= ~STATE_DECODE_QUEUED;
	article_add_flag (dd->sdata, dd->gdata, adata,
		(success ? STATE_DECODED : STATE_DECODE_FAILED), TRUE);
}

/***
**** Private Routines
***/

/**
 * This is the function that the decode thread runs.
 */

static void
decode_open (const char *fname)
{
	const char *mime_type = NULL;
	const char *mime_prog = NULL;
	char *quoted_fname = NULL;
	char *exec = NULL;

	if (!((mime_type = gnome_mime_type (fname))))
		return;
	if (!((mime_prog = gnome_mime_program (mime_type))))
		return;
	if (!strstr (mime_prog, "%f"))
		return;

	quoted_fname = g_strdup_printf ("\"%s\"", fname);
	exec = pan_substitute (mime_prog, "%f", quoted_fname);
	g_free (quoted_fname);
	gnome_execute_shell (NULL, exec);
	g_free (exec);
}

/***************************************************************
* UUdeview Lib Stuff 
***************************************************************/

static void
UUD_message_callback (void *opaque, char *msg, int level)
{
	debug(DEBUG_DECODE,"UUDeview says: '%s' (level %d)", msg, level);

	if ((level == UUMSG_WARNING) || (level == UUMSG_ERROR)) {
		pan_error_dialog (_("An error occured during decoding.\n%s"), msg);
	}
}


static int
UUD_busy_callback (void *opaque, uuprogress *progress)
{
	StatusItem *item = STATUS_ITEM(opaque);

	const int i = (progress->percent + 100.0 * (progress->partno-1))
		/ progress->numparts;
	if ( 0<=i && i<=100 )
		status_item_emit_progress (item, i);

	return 0;
}


static char *
UUD_file_name_filter (void *opaque, char *fname)
{
	char *ptr;

	if (fname == NULL)
		return NULL;

	/*
	 * strip directory information
	 */

	if ((ptr = _FP_strrchr(fname, '/')) != NULL)
		return ptr + 1;
	else if ((ptr = _FP_strrchr(fname, '\\')) != NULL)
		return ptr + 1;

	/* increment UNKNOWN.??? if necessary */
	if (strncmp(fname, "UNKNOWN", 7) == 0) {
		struct stat finfo;
		int fnum;

		if (lstat(fname, &finfo) == -1)
			return fname;

		do {
			fnum = atoi (&fname[8]);
			fnum++;
			sprintf(&fname[8], "%.3u", fnum);
		} while (lstat(fname, &finfo) == 0);
	}

	return fname;
}
