/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "acache.h"
#include "article-db.h"
#include "date.h"
#include "globals.h"
#include "log.h"
#include "nntp.h"
#include "queue.h"

#include "util.h"

gboolean online_prompt = TRUE;

extern const char* sockwrite_err_msg;
extern const char* sockread_err_msg;

void
nntp_disconnect (PanSocket *sock)
{
	pan_socket_putline (sock, "QUIT\r\n");
}

const gchar*
nntp_login (PanSocket *sock, Server* sdata)
{
	int ival;
	const gchar* buffer;

	g_return_val_if_fail (sock!=NULL, NULL);

	if (pan_socket_getline (sock, &buffer))
		return _("Error reading from socket.");

	/* attempt authentication.  This isn't implemented exactly as
	 * "Common NNTP Extensions" specifies, being that we didn't
	 * wait to be asked for it, but hopefully it will work.
	 */
	if (sdata->need_auth)
	{
		if (pan_socket_putline_va (sock, "AUTHINFO user %s\r\n", sdata->username))
			return _("Error writing to socket.");
		if (pan_socket_getline (sock, &buffer))
			return _("Error reading from socket.");
		if (atoi(buffer) == 381)
		{
			/* 381 Means: More authentication information needed */
			if (pan_socket_putline_va (sock, "AUTHINFO pass %s\r\n", sdata->password))
				return _("Error writing to socket.");

			if (pan_socket_getline (sock, &buffer))
				return _("Error reading from socket.");

			/* 281 Means: Authentication accepted */
			if (atoi(buffer) != 281) {
				log_add_va ( _("Authentication for server '%s' unsuccessful."),
					sdata->name);
				return buffer;
			}
		}
		/* Else authentication not required or not understood */
	}

	if (pan_socket_putline (sock, "MODE READER\r\n"))
		return _("Error writing to socket.");
	if (pan_socket_getline (sock, &buffer))
		return _("Error reading from socket.");

	ival = atoi(buffer);
	if (ival == 200)
		sdata->posting = TRUE;
	else if (ival == 201)
		sdata->posting = FALSE;
	else if (ival == 500) /* command not recognized; "mode reader" is an extension */
		sdata->posting = TRUE; /* a wild guess */
	else
		return buffer;

	return NULL;
}




int
nntp_set_group (PanSocket *sock,
                const Group *data)
{
	const gchar *buffer = NULL;

	g_return_val_if_fail (data, -1);

	/* write the request */
	if (pan_socket_putline_va (sock, "GROUP %s\r\n", data->name))
		return -1;

	/* get the result string */
	if (pan_socket_getline (sock, &buffer))
		return -1;

	/* interpret the result string */
	if (atoi(buffer)!=210 && atoi(buffer)!=211)
		return -1;

	return 0;
}



/**
 * nntp_post:
 * @msg: ???
 *
 * Return value:
 **/
int
nntp_post (PanSocket *sock,
           const gchar *msg)
{
	const gchar *buffer;

	if (Pan.mute) {
		g_message ("Mute: Your Mesage will not actually be posted.");
		printf ("\n\n\nYour message\n%s<end of message>\n\n\n", msg);
		fflush (NULL);
		return 0;
	}

	if (pan_socket_putline (sock, "POST\r\n")) {
		pan_error_dialog (_("Error writing to socket during post."));
		return -1;
	}

	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during post."));
		return -1;
	}

	if (atoi(buffer) == 340) { /* 340 Means: Send Article To Be Posted */
		if (pan_socket_putline_va (sock, "%s\r\n.\r\n", msg)) {
			pan_error_dialog (_("Error writing to socket during post."));
			return -1;
		}
		if (pan_socket_getline (sock, &buffer)) {
			pan_error_dialog (_("Error reading from socket during post."));
			return -1;
		}
		if (atoi(buffer) == 240) { /* 240 Means: Article Posted OK */
			return 0;
		}
	}

	pan_error_dialog (_("Error posting message.  Server said:\n'%s'"), buffer);
	return -1;
}

/*--------------------------------------------------------------------
 * download and decode an article
 *--------------------------------------------------------------------*/

gint
nntp_send_noop (PanSocket *sock)
{
	const char *buffer;

	if (pan_socket_putline (sock, "HELP\r\n")) {
		pan_error_dialog (_("Error writing to socket during HELP no-op."));
		return -1;
	}
	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during HELP no-op."));
		return -1;
	}
	if (atoi(buffer) == 100) { /* 100 = help text follows */
 		while (strncmp(buffer, ".\r\n", 3)) {
			if (pan_socket_getline (sock, &buffer)) {
				pan_error_dialog (_("Error reading from socket during HELP no-op."));
				return -1;
			}
		}
	}

	return 0;
}


int
nntp_article_download (
	StatusItem* item,
	PanSocket* sock,
	const Server* sdata,
	const Group* gdata,
	Article* adata,
	const gboolean* abort,
	gboolean verbose)
{
	gboolean adata_headers_changed = FALSE;
	const char *buffer;
	FILE *f;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);

	if (verbose)
		status_item_emit_status (
			STATUS_ITEM(item),
			_("Downloading article '%s'"),
			adata->subject);

#if 0
	/* set group */
	if (*abort || nntp_set_group(sock, gdata))
	{
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't select group '%s'"), gdata->name);
		return -1;
	}
#endif

	/* get the head */
	if (*abort || pan_socket_putline_va (sock, "HEAD %s\r\n", adata->number))
	{
		status_item_emit_error (STATUS_ITEM(item), sockwrite_err_msg);
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer))
	{
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	if (atoi(buffer) != 221)
	{
		char* pch = g_strdup (buffer);
		pch[strcspn(pch,"\r\n")] = '\0';
		status_item_emit_error(STATUS_ITEM(item), _("NNTP \"HEAD\" command returned `%s'."), pch);
		g_free (pch);
		return -1;
	}
	if (*abort || pan_socket_getline (sock,&buffer))
	{
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}

	f = acache_open_header (adata);
	while (strncmp(buffer, ".\r\n", 3))
	{
		char* pch = NULL;
		const char* name = NULL;

		/* check for exit... */
		if (*abort) {
			acache_close_header (f);
			acache_delete (adata->message_id);
			return -1;
		}

		/* write to acache... */
		acache_putline_header (f, buffer);

		/* parse it for adata... */
		if ((pch = strrchr (buffer, '\x0D'))) /* trim linefeed */
			*pch = '\0';
		name = "Newsgroups: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			g_free (adata->newsgroups);
			adata->newsgroups = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}
		name = "Reply-To: ";
		if (!strncmp (buffer, name, strlen(name))) {
			g_free (adata->reply_to);
			adata->reply_to = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}
		name = "Followup-To: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			g_free (adata->followup_to);
			adata->followup_to = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}

		/* get the next line... */
		if (*abort || pan_socket_getline (sock, &buffer)) {
			status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
			acache_close_header (f);
			acache_delete (adata->message_id);
			return -1;
		}
	}
	acache_close_header (f);

	if (adata_headers_changed)
		article_update (sdata, gdata, adata, TRUE);

	if (*abort || pan_socket_putline_va (sock, "BODY %s\r\n", adata->number)) {
		status_item_emit_error (STATUS_ITEM(item), sockwrite_err_msg);
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	if (atoi(buffer) != 222) {
		return -1;
	}
	
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	f = acache_open_body (adata);

	if (verbose)
		status_item_emit_init_steps (STATUS_ITEM(item), adata->linecount);

	while (strncmp(buffer, ".\r\n", 3))
	{
		if (*abort)
		{
			acache_close_body (f);
			acache_delete (adata->message_id);
			return -1;
		}
		acache_putline_body (f, buffer);

		if (verbose)
			status_item_emit_next_step(STATUS_ITEM(item));

		if (*abort || pan_socket_getline (sock, &buffer))
		{
			status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
			acache_close_body (f);
			acache_delete (adata->message_id);
			return -1;
		}
	}
	acache_close_body (f);
	acache_file_checkin (adata->message_id);

	if (verbose)
		status_item_emit_progress(STATUS_ITEM(item), 100);

	return 0;
}


int
nntp_get_Group (
	StatusItem* status,
	PanSocket* sock,
	const char* group_name,
	int* article_qty,
	int* low_num,
	int* high_num,
	const gboolean* abort )
{
	const char *buffer = NULL;

	/* change to the right group */
	if (*abort || pan_socket_putline_va (sock, "GROUP %s\r\n", group_name))
	{
		status_item_emit_error (status, sockwrite_err_msg);
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer))
	{
		status_item_emit_error (status, sockread_err_msg);
		return -1;
	}
	if (strtol(buffer, NULL, 10) != 211)
	{
		status_item_emit_error (status, _("(Couldn't select group '%s'"), group_name);
		return -1;
	}
	sscanf (buffer, "%*d %d %d %d", article_qty, low_num, high_num);
	return 0;
}

