/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *
 *
 * A note on implementation
 * The queue defers all public tasks (like insert, remove, etc.)
 * to the queue thread.  This ensures that all tasks are handled
 * in sequence, rather than concurrently.  This ensures that callbacks
 * will be invoked in sync with the queue, so the queue's state won't
 * change during the callback.
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <pthread.h>
#include <sys/time.h>
#include <errno.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "debug.h"
#include "log.h"
#include "nntp.h"
#include "queue.h"
#include "gui.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

typedef enum
{
	TODO_INSERT,
	TODO_REMOVE,
	TODO_MOVE,
	TODO_REQUEUE,
	TODO_TASK_ABORT,
	TODO_SET_ONLINE,
	TODO_SET_MAX_SOCKET_QTY,
	TODO_SET_MAX_TRIES,
	TODO_DEDICATE_THREAD
}
QueueTodoAction;

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

typedef struct
{
	Task *task;
	int index_1;
	int index_2;
	QueueTodoAction action;
}
QueueTodo;

typedef struct
{
	PanSocket *socket;
	Server *server;
	Task *task;	/* NULL if idle */
}
SocketInfo;

/*********************
**********************  Private Function Prototypes
*********************/

static void queue_mainloop (void);
static void queue_run_what_we_can (void);
static void queue_set_task_status (Task* task, QueueTaskStatus status);
static void queue_wakeup (void);
static gint sockets_find_socket (gconstpointer a, gconstpointer b);
static void socket_cleanup (Task* task);
static void queue_new_todo (QueueTodoAction action, Task* task, int i1, int i2);
static gint queue_get_length (void);
static void queue_do_todo (void);

/*********************
**********************  Constants
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

PanCallback *queue_task_added = NULL;
PanCallback *queue_task_removed = NULL;
PanCallback *queue_task_moved = NULL;
PanCallback *queue_task_status_changed = NULL;
PanCallback *queue_online_status_changed = NULL;
PanCallback *queue_max_tries_changed = NULL;

pthread_mutex_t qlock;

/***********
************  Private
***********/

static GSList* todo_list = NULL;
static GSList* queue_list = NULL;

static pthread_mutex_t todo_lock;
static pthread_mutex_t slock;
static pthread_mutex_t qcond_m;
static pthread_cond_t qcond = PTHREAD_COND_INITIALIZER;

static pthread_t queue_thread = 0;

static GHashTable* task_to_status = NULL;

static GSList * sockets = NULL;

static int max_socket_qty = 4;

static int max_tries = 5;

static gboolean dedicate_one_thread_to_reading = TRUE;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

static void
fire_online_status_changed (void)
{
	int args[3];
	args[0] = max_socket_qty>0;
	args[1] = max_socket_qty;
	args[2] = dedicate_one_thread_to_reading;
	pan_callback_call (queue_online_status_changed, NULL, args);
}


void
queue_init (int tries, int socket_qty)
{
	task_to_status = g_hash_table_new (g_direct_hash, g_direct_equal);

	queue_task_added = pan_callback_new();
	queue_task_removed = pan_callback_new();
	queue_task_moved = pan_callback_new();
	queue_task_status_changed = pan_callback_new();
	queue_online_status_changed = pan_callback_new();
	queue_max_tries_changed = pan_callback_new();

	pthread_mutex_init (&slock, NULL);
	pthread_mutex_init (&qlock, NULL);
	pthread_mutex_init (&qcond_m, NULL);
	pthread_mutex_init (&todo_lock, NULL);

	max_tries = tries;
	max_socket_qty = socket_qty;

	/* fire up the queue thread */
	pthread_create (&queue_thread, NULL, (void*)queue_mainloop, NULL);
	pthread_detach (queue_thread);
}

void
queue_add (Task* task)
{
	queue_task_insert (task, task->high_priority ? 0 : -1);
}

const GSList*
queue_get_tasks (void)
{
	return queue_list;
}

/**
***
**/

gboolean
queue_has_dedicated_article_thread (void)
{
	return dedicate_one_thread_to_reading;
}
void
queue_set_dedicated_article_thread (gboolean b)
{
	queue_new_todo (TODO_DEDICATE_THREAD, NULL, b?1:0, -1);
}
static void
real_queue_set_dedicated_article_thread (gboolean b)
{
	if (b != dedicate_one_thread_to_reading)
	{
		dedicate_one_thread_to_reading = b;
		fire_online_status_changed ();
		queue_wakeup();
	}
}

/**
***
**/

void
queue_task_abort (Task* task)
{
	queue_new_todo (TODO_TASK_ABORT, task, -1, -1);
}
static void real_queue_task_abort (Task* task)
{
	task_hint_abort (TASK(task), TRUE);
	queue_set_task_status (task, QUEUE_TASK_STATUS_ABORTING);
}

/**
***
**/

void
queue_task_insert (Task* task, int index)
{
	queue_new_todo (TODO_INSERT, task, index, -1);
}
static void
real_queue_task_insert (Task* task, int index)
{
	debug (DEBUG_QUEUE,
		"%s:%d queue_task_insert %p to pos %d",
		__FILE__, __LINE__, task, index);

	g_return_if_fail (task!=NULL);
	g_return_if_fail (index>=-1);
	g_return_if_fail (index<=(int)queue_get_length());

	pan_object_ref (PAN_OBJECT(task));
	pan_object_sink (PAN_OBJECT(task));

	/* add to the list */
	pthread_mutex_lock (&qlock);
	queue_list = g_slist_insert (queue_list, task, index);
	g_hash_table_insert (task_to_status, task,
		GINT_TO_POINTER(QUEUE_TASK_STATUS_QUEUED));
	pthread_mutex_unlock (&qlock);
	gui_set_queue_size ((int)g_slist_length(queue_list));

	/* call callbacks */
	debug (DEBUG_QUEUE,
		"%s:%d queue size is now %d, calling callbacks",
		__FILE__, __LINE__, queue_get_length());
	pan_callback_call (queue_task_added, task, GINT_TO_POINTER(index));

	queue_wakeup ();
	
	debug (DEBUG_QUEUE,
		"%s:%d queue_task_insert %p to pos %d done",
		__FILE__, __LINE__, task, index);
}

/**
***
**/

void
queue_task_remove (Task* task)
{
	queue_new_todo (TODO_REMOVE, task, -1, -1);
}
static void
real_queue_task_remove (Task* task)
{
	debug (DEBUG_QUEUE,
		"%s:%d queue_task_remove %p",
		__FILE__, __LINE__, task);

	g_return_if_fail (task!=NULL);

	if (task->thread_id != 0) /* if it's running, send an abort message */
	{
		task_hint_abort (task, TRUE);
		queue_set_task_status (task, QUEUE_TASK_STATUS_ABORTING);
	}
	else /* not running; simple cleanup. queue_run_thread() zeroes out thread_id */
	{
		const gboolean had_socket = task->sock != NULL;
		socket_cleanup (task);

		/* remove the task */
		pthread_mutex_lock (&qlock);
		queue_list = g_slist_remove (queue_list, task);
		pthread_mutex_unlock (&qlock);
		gui_set_queue_size ((int)g_slist_length(queue_list));

		pan_callback_call (queue_task_removed, task, NULL);
		g_hash_table_remove (task_to_status, task);
		pan_object_unref (PAN_OBJECT(task));

		if (had_socket) /* maybe another socket task can run now */
			queue_wakeup ();
	}

	debug (DEBUG_QUEUE,
		"%s:%d queue_task_remove %p done",
		__FILE__, __LINE__, task);
}

/**
***
**/

void
queue_task_move (Task* task, int index)
{
	queue_new_todo (TODO_MOVE, task, index, -1);
}
static void
real_queue_task_move (Task* task, int index)
{
	int old_index;
	int indices [2];

	debug (DEBUG_QUEUE,
		"%s:%d queue_task_move %p to %d",
		__FILE__, __LINE__, task, index);

	g_return_if_fail (task!=NULL);
	g_return_if_fail (index>=-1);
	g_return_if_fail (index<=(int)queue_get_length());
	
	pthread_mutex_lock (&qlock);
	
	old_index = g_slist_index (queue_list, task);
	g_return_if_fail (old_index!=-1);

	queue_list = g_slist_remove (queue_list, task);
	queue_list = g_slist_insert (queue_list, task, index);
	
	indices[0] = old_index;
	indices[1] = index;

	pthread_mutex_unlock (&qlock);

	debug (DEBUG_QUEUE,
		"%s:%d queue_task_move %p calling callbacks; "
		"old index is %d, new index is %d",
		__FILE__, __LINE__, task, old_index, index);
	pan_callback_call (queue_task_moved, task, indices);

	debug (DEBUG_QUEUE,
		"%s:%d queue_task_move %p done",
		__FILE__, __LINE__, task);
}

/**
***
**/

void
queue_task_requeue_failed (Task* task, int index)
{
	queue_new_todo (TODO_REQUEUE, task, index, -1);
}
static void
real_queue_task_requeue_failed (Task* task, int index)
{
	task->tries = 0;
	queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
}

/**
***
**/

int
queue_get_max_sockets (void)
{
	return max_socket_qty;
}

void
queue_set_max_sockets (int i)
{
	queue_new_todo (TODO_SET_MAX_SOCKET_QTY, NULL, i, -1);
}
static void
real_queue_set_max_sockets (int i)
{
	if (max_socket_qty != i)
	{
		max_socket_qty = i;
		fire_online_status_changed ();
	}
}

/**
***
**/

gboolean
queue_is_online (void)
{
	return max_socket_qty>0;
}

void
queue_set_online (gboolean b)
{
	queue_new_todo (TODO_SET_ONLINE, NULL, b?1:0, -1);
}
static void
real_queue_set_online (gboolean b)
{
	const int old = max_socket_qty;

	if (queue_is_online()!=b)
	{
		max_socket_qty = b ? abs(old) : -abs(old);
		fire_online_status_changed ();

		if (max_socket_qty>old)
			queue_wakeup();
	}
}

/**
***
**/

int
queue_get_max_tries (void)
{
	return max_tries;
}

void
queue_set_max_tries (int i)
{
	queue_new_todo (TODO_SET_MAX_TRIES, NULL, i, -1);
}
static void
real_queue_set_max_tries (int i)
{
	max_tries = i;
	pan_callback_call (queue_max_tries_changed, NULL, GINT_TO_POINTER(i));
}

/**
***
**/

QueueTaskStatus
queue_get_task_status (const Task* task)
{
	QueueTaskStatus status = GPOINTER_TO_INT (
		g_hash_table_lookup (task_to_status, task));
	if (!status)
		status = QUEUE_TASK_STATUS_NOT_QUEUED;
	return status;
}

static void
queue_set_task_status (Task* task, QueueTaskStatus status)
{
	g_return_if_fail (task!=NULL);
	g_return_if_fail (g_slist_index(queue_list, task) != -1);

	g_hash_table_insert (
		task_to_status, task,
		GINT_TO_POINTER(status));

	pan_callback_call (
		queue_task_status_changed, task,
		GINT_TO_POINTER(status)); 
}

/************
*************  PRIVATE ROUTINES
************/

static void
queue_new_todo (QueueTodoAction action, Task* task, int i1, int i2)
{
	QueueTodo *todo = g_new0 (QueueTodo, 1);
	todo->task = task;
	todo->index_1 = i1;
	todo->index_2 = i2;
	todo->action = action;

	pthread_mutex_lock (&todo_lock);
	todo_list = g_slist_append (todo_list, todo);
	debug (DEBUG_QUEUE, "%s:%d todo queue now has %d tasks (new task type: %d, int 1: %d)",
		__FILE__, __LINE__, (int)g_slist_length(todo_list), action, i1);
	pthread_mutex_unlock (&todo_lock);

	queue_wakeup ();
}

static void
queue_wakeup (void)
{
        pthread_mutex_lock (&qcond_m);
        pthread_cond_broadcast (&qcond);
        pthread_mutex_unlock (&qcond_m);
}

static gint
queue_get_length (void)
{
	gint size = -1;
	pthread_mutex_lock (&qlock);
	size = g_slist_length (queue_list);
	pthread_mutex_unlock (&qlock);
	return size;
}

static gint
sockets_find_socket (gconstpointer a, gconstpointer b)
{
	const SocketInfo * info = (const SocketInfo*)a;
	const PanSocket * socket = (const PanSocket*)b;
	return info->socket != socket;
}

static void
queue_run_thread (void* vp)
{
	/* preparing to run */
	int retval=0;
	Task* task = TASK(vp);
	gchar* desc = status_item_describe (STATUS_ITEM(task));
	gboolean aborted;

	debug (DEBUG_QUEUE,
		"%s:%d queue_run_thread: task %p [%s]",
		__FILE__, __LINE__, task, desc);

	/* run */
	gui_add_status_item (STATUS_ITEM(task));
	if (task->sock)
		pan_socket_reset_statistics (task->sock);
	retval = task_run (task);
	debug (DEBUG_QUEUE,
		"%s:%d queue task[%p] (%s) returned: %d",
		__FILE__, __LINE__, task,desc,retval);
	g_free(desc);

	/* clean up gui */
	gui_remove_status_item (STATUS_ITEM(task));

	/* clean up socket */
	aborted = queue_get_task_status(task) == QUEUE_TASK_STATUS_ABORTING;
	if (aborted)
		task->sock->error = TRUE;
	socket_cleanup (task);

	/* clean up the task */
	task->thread_id = 0;
	if (!retval) /* success */
	{
		queue_task_remove (task);
	}
	else if (task->tries<max_tries && !aborted) /* retry */
	{
		++task->tries;
		queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
	}
	else /* aborted or failed */
	{
		task->tries = 0;
		queue_task_abort (task);
	}

	debug (DEBUG_QUEUE,
		"%s:%d queue_run_thread: task %p done",
		__FILE__, __LINE__, vp);

	pthread_exit (NULL);
}


static void
queue_run (Task* task)
{
	g_return_if_fail (task!=NULL);

	debug (DEBUG_QUEUE,
		"%s:%d starting a thread to run task %p, sock %p",
		__FILE__, __LINE__, task, task->sock );

	queue_set_task_status (task, QUEUE_TASK_STATUS_RUNNING);

	pthread_create (&task->thread_id, NULL, (void*)queue_run_thread, task);
	pthread_detach (task->thread_id);
}


/**
 * Return -1 if we can't connect because all lines are busy
 * Return -2 if we can't connect because of a connection failure
 */
static gint
queue_get_socket_for_task (Task *task)
{
	GSList* sl = NULL;
	gint retval = -1;
	const gboolean is_reader_task = TASK(task)->is_article_reader;
	gboolean skip_first_idle =
		dedicate_one_thread_to_reading && !is_reader_task;
	gboolean reader_socket_exists = FALSE;
	Server *sdata = task->sdata;
	int socket_qty;

	debug (DEBUG_QUEUE,
		"%s:%d queue_get_socket_for_task %p",
		__FILE__, __LINE__, task );

	pthread_mutex_lock (&slock);

	/* if an idle socket for that server, use it. */
	socket_qty = 0;
	for (sl=sockets; sl!=NULL && retval!=0; sl=sl->next)
	{
		SocketInfo * si = (SocketInfo*) sl->data;

		if (si->server != task->sdata) /* ignore other servers */
			continue;

		++socket_qty;

		if (si->task != NULL) /* socket already in use */
			continue;

		if (skip_first_idle) { /* save one article reader socket */
			reader_socket_exists = TRUE;
			skip_first_idle = FALSE;
			continue;
		}

		si->task = task;
		task->sock = si->socket;

		debug (DEBUG_QUEUE,
			"%s:%d queue_get_socket_for_task %p "
			"reusing socket #%d (%p)",
			__FILE__, __LINE__, 
			task, g_slist_index(sockets, si), si->socket);

		retval = 0;
	}

	/* if no idle sockets for that server... */
	if (retval && socket_qty<max_socket_qty)
	{
		const int current_connections = socket_qty;
		int slots_open = sdata->max_connections - current_connections;

		debug (DEBUG_QUEUE,
			"%s:%d queue_get_socket_for_task %p "
			"no idle sockets", __FILE__, __LINE__ );

		/* leave an article reader socket free */
		if (!reader_socket_exists
			&& dedicate_one_thread_to_reading
			&& !is_reader_task
			&& max_socket_qty>1
			&& sdata->max_connections>1)
				--slots_open;

		if (slots_open <= 0)
		{
			debug (DEBUG_QUEUE,
				"%s:%d we've already got %d connections "
				"and %d reader connections",
				__FILE__, __LINE__,
				current_connections,
				(reader_socket_exists?1:0) );

			retval = -1;
		}
		else	
		{
			const char* errmsg = NULL;

			/* open a new socket.. */
			PanSocket *sock = NULL;
			debug (DEBUG_QUEUE,
				"%s:%d making a new socket to %s",
				__FILE__, __LINE__, 
				sdata->address);

			sock = pan_socket_new (sdata->address, sdata->port);
			pan_object_ref (PAN_OBJECT(sock));
			pan_object_sink (PAN_OBJECT(sock));

			debug (DEBUG_QUEUE,
				"%s:%d new socket %p",
				__FILE__, __LINE__, sock);
			if (sock->sockfd == 0)
				errmsg = "Unable to connect to server";

			/* try to log in... */
			if (!errmsg)
				errmsg = nntp_login(sock, sdata);
			if (errmsg!=NULL)
			{
				log_add(errmsg);
				g_warning(errmsg);
				if (sock)
					pan_object_unref (PAN_OBJECT(sock));
				sock = NULL;

				debug (DEBUG_QUEUE,
					"%s:%d %p can't log in!",
					__FILE__, __LINE__, sock);

				retval = -2;
			}
			else
			{
				SocketInfo * si = NULL;

				debug (DEBUG_QUEUE,
					"%s:%d %p logged in",
					__FILE__, __LINE__, sock);

				/* looks like we got a new socket okay... */
				si = g_new0 (SocketInfo, 1);
				si->socket = sock;
				si->server = task->sdata;
				si->task = task;
				task->sock = sock;
				sockets = g_slist_prepend (sockets, si);
				retval = 0;
			}
		}
	}

	pthread_mutex_unlock (&slock);

	debug (DEBUG_QUEUE,
		"%s:%d queue_get_socket_for_item was %ssuccessful",
		__FILE__, __LINE__, (retval?"not ":""));

	return retval;
}

static void
queue_run_what_we_can (void)
{
	GSList *l = NULL;

	debug (DEBUG_QUEUE,
		"%s:%d queue_run_what_we_can (%d in queue)",
		__FILE__, __LINE__,
		queue_get_length());

	/* walk through the queue and see if
	   any tasks are waiting to be run. */
	l = queue_list;
	while (l!=NULL)
	{
		Task *task = TASK(l->data);

		/* if it's not waiting to be run, leave it alone */
		if (queue_get_task_status(task) != QUEUE_TASK_STATUS_QUEUED)
		{
			debug (DEBUG_QUEUE,
				"%s:%d task isn't waiting to be run",
				__FILE__, __LINE__);

		}

		/* if it doesn't need a socket, no problemo! */
		else if (!task->needs_socket)
		{
			debug (DEBUG_QUEUE,
				"%s:%d task %d needs no socket; running it now",
				__FILE__, __LINE__, task);

			queue_run (task);
		}

		/* try to get a socket... */
		else if (max_socket_qty>0)
		{
			int i;
			queue_set_task_status (task, QUEUE_TASK_STATUS_CONNECTING);
			i = queue_get_socket_for_task (task);
			if (i==0)
			{
				debug (DEBUG_QUEUE,
					"%s:%d task %d got a socket!",
					__FILE__, __LINE__, task);

				queue_run (task);
			}
			else
			{
				if (i==-2) /* connection failure */
				{
					status_item_emit_error (STATUS_ITEM(task), _("Connect Failure"));

					++task->tries;
					if (task->tries > max_tries)
						queue_task_abort (task);
				}

				queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
			}
		}

		if (todo_list)
		{
				queue_do_todo ();
				l = queue_list;
		}
		else
		{
				l = l->next;
		}
	}

	debug (DEBUG_QUEUE,
		"%s:%d queue_run_what_we_can done",
		__FILE__, __LINE__);
}

/**
 * Destroy or recycle a socket when a task is done with it.
 * @param task the finished task.
 */
static void
socket_cleanup (Task* task)
{
	debug (DEBUG_QUEUE,
		"%s:%d socket_cleanup for task %p (sock %p)",
		__FILE__, __LINE__, task, task?task->sock:NULL);

	if (task && task->sock)
	{
		PanSocket* socket = task->sock;
		GSList* l = NULL;
		SocketInfo * info = NULL;
		int server_socket_qty;

		pthread_mutex_lock (&slock);

		/* count the number of sockets this server has */
		server_socket_qty = 0;
		for (l=sockets; l!=NULL; l=l->next)
			if (((SocketInfo*)l->data)->server == task->sdata)
				++server_socket_qty;

		/* find this SocketInfo in our tables */
		l = g_slist_find_custom (sockets, socket, sockets_find_socket);
		g_assert (l!=NULL);
		info = (SocketInfo*) l->data;

		if (!socket->error
			&& (int)g_slist_length(sockets)<=max_socket_qty
			&& server_socket_qty<=task->sdata->max_connections)
		{
		       	/* if socket's okay, reuse it */
			info->task = NULL;

			debug (DEBUG_QUEUE,
				"%s:%d socket_cleanup: recycling socket %p",
				__FILE__, __LINE__, socket);
		}
		else
		{
			/* otherwise throw it away */
			pan_object_unref (PAN_OBJECT(socket));
			sockets = g_slist_remove (sockets, info);
			g_free (info);

			debug (DEBUG_QUEUE,
				"%s:%d socket_cleanup: throwing away socket %p",
				__FILE__, __LINE__, socket);
		}

		task->sock = NULL;

		pthread_mutex_unlock (&slock);
	}
}

static void
sockets_upkeep (void)
{
	GSList* l = NULL;
	GSList* cull = NULL;
	int socket_qty;

	pthread_mutex_lock (&slock);

	socket_qty = g_slist_length (sockets);

	debug (DEBUG_QUEUE,
		"%s:%d socket_upkeep -- %d sockets",
		__FILE__, __LINE__, socket_qty);

	for (l=sockets; l!=NULL; l=l->next)
	{
		const time_t current_time = time(0);
		SocketInfo * info;
		PanSocket * sock;
		int age;
		gboolean destroy;

		info = (SocketInfo*) l->data;
		g_assert (info != NULL);
		sock = info->socket;
		g_assert (sock != NULL);
		age = current_time - sock->last_action_time;
		destroy = ((info->task==NULL)
		       	&& ((age > info->server->idle_secs_before_timeout)
				|| socket_qty > max_socket_qty));

		/* if it's idle and we're keeping it, send a keepalive msg.
		   Note that we don't let this affect the socket's
		   age, otherwise they would never grow old and close... */
		if (info->task==NULL && !destroy) {
			const time_t last_action_time = sock->last_action_time;
			debug (DEBUG_QUEUE, "sending %p a keepalive", sock);
			destroy = nntp_send_noop(sock) != 0;
			sock->last_action_time = last_action_time;
		}

		/* either keep the socket or cull it from the herd */
		if (destroy) {
			--socket_qty;
			cull = g_slist_prepend (cull, info);
		}
	}

	debug (DEBUG_QUEUE,
		"%s:%d sockets_upkeep: culling %d sockets",
		__FILE__, __LINE__, g_slist_length(cull));

	for (l=cull; l!=NULL; l=l->next)
	{
		const time_t current_time = time(0);
		SocketInfo * info = (SocketInfo*) l->data;
		PanSocket * sock = info->socket;
		const int age = current_time - sock->last_action_time;

		debug (DEBUG_QUEUE,
			"%s:%d socket %p idle for %d seconds.. closing",
			__FILE__, __LINE__, sock, age);

		nntp_disconnect (sock);
		pan_object_unref (PAN_OBJECT(sock));
		sockets = g_slist_remove (sockets, info);
		g_free (info);
	}

	pan_warn_if_fail (g_slist_length(sockets) == socket_qty);

	pthread_mutex_unlock (&slock);

	g_slist_free (cull);

	debug (DEBUG_QUEUE,
		"%s:%d socket_upkeep done",
		__FILE__, __LINE__);
}

static void
queue_do_todo (void)
{
	GSList * list = NULL;
	GSList * l = NULL;

	pthread_mutex_lock (&todo_lock);
	list = todo_list;
	todo_list = NULL;
	pthread_mutex_unlock (&todo_lock);

	for (l=list; l!=NULL; l=l->next)
	{
		QueueTodo* a = (QueueTodo*) l->data;
		switch (a->action)
		{
			case TODO_INSERT:
				real_queue_task_insert (a->task, a->index_1);
				break;
			case TODO_REMOVE:
				real_queue_task_remove (a->task);
				break;
			case TODO_MOVE:
				real_queue_task_move (a->task, a->index_1);
				break;
			case TODO_REQUEUE:
				real_queue_task_requeue_failed (
					a->task, a->index_1);
				break;
			case TODO_TASK_ABORT:
				real_queue_task_abort (a->task);
				break;
			case TODO_SET_ONLINE:
				real_queue_set_online (a->index_1!=0);
				break;
			case TODO_SET_MAX_SOCKET_QTY:
				real_queue_set_max_sockets (a->index_1);
				break;
			case TODO_SET_MAX_TRIES:
				real_queue_set_max_tries (a->index_1);
				break;
			case TODO_DEDICATE_THREAD:
				real_queue_set_dedicated_article_thread (
					a->index_1!=0);
				break;
			default:
				pan_warn_if_reached();
				break;
		}
	}

	/* throw the todo list away */
	g_slist_foreach (list, (GFunc)g_free, NULL);
	g_slist_free (list);
	list = NULL;
}

static void
queue_mainloop (void)
{
	const int TIMEOUT_SECS = 20;
	struct timeval now;
	struct timespec timeout;
	int retcode = 0;

	for (;;)
        {
		pthread_mutex_lock (&qcond_m);
		gettimeofday (&now, NULL);

                /* Wait TIMEOUT_SECS for something to hit the queue */
                timeout.tv_sec = now.tv_sec + TIMEOUT_SECS;
                timeout.tv_nsec = timeout.tv_sec;
                retcode = pthread_cond_timedwait (&qcond, &qcond_m, &timeout);
                pthread_mutex_unlock (&qcond_m);

		/* upkeep */
		queue_do_todo ();

                /* process any queue items */
                queue_run_what_we_can ();

                /* if timeout, do socket upkeep */
                if (retcode==ETIMEDOUT) {
			sockets_upkeep ();
                }
	}
}
