/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-config.h>
#include <libgnome/gnome-i18n.h>

#include "globals.h"
#include "grouplist.h"
#include "gui.h"
#include "prefs.h"
#include "util.h"

static void server_list_save (void);


void
server_shutdown (void)
{
	GSList *l;
	Server *last;
	for (l = Pan.serverlist; l; l = l->next) {
		Server *sdata = l->data;
		pan_db_close (sdata->db);
	}
	last = grouplist_get_current_server ();
	if (last && last->name) {
		gnome_config_set_string ("/Pan/State/Server", last->name);
		gnome_config_sync ();
	}
}

static void
server_selected_cb (GtkWidget *w, Server *server)
{
	if (GTK_CHECK_MENU_ITEM(w)->active)
		grouplist_server_selected (server);
}

static GnomeUIInfo end = GNOMEUIINFO_END;


static int server_menu_len = 0;
static GnomeUIInfo * server_menu = NULL;
static GnomeUIInfo * server_menu_radiobox = NULL;

void
server_menu_update (void)
{
	GSList* l;
	GnomeUIInfo * buf;
	GnomeUIInfo * radio_box;
	GString * str = g_string_new (NULL);
	int i;

	/* build the new menu */
	l = Pan.serverlist;
	buf = g_new0 (GnomeUIInfo, g_slist_length(l)+2);
	i = 0;
	if (!l)
	{
		GnomeUIInfo * info = &buf[i];
		info->type = GNOME_APP_UI_ITEM;
		info->label = _("No Servers Defined");
		info->hint = _("No Servers Defined");
		info->moreinfo = NULL;
		info->user_data = NULL;
		++i;
	}
	for (; l!=NULL; l=l->next, ++i)
	{
		Server *s = SERVER(l->data);
		GnomeUIInfo * info = &buf[i];

		info->type = GNOME_APP_UI_ITEM;
		info->label = s->name;
		info->hint = s->name;
		info->moreinfo = server_selected_cb;
		info->user_data = s;

		/* open the server's db... */
		if (s->db == NULL)
		{
			g_string_sprintf(str,"/%s/%s.db",data_dir,s->name);
			s->db = pan_db_open (str->str);
		}
	}
	buf[i] = end;
	radio_box = g_new0 (GnomeUIInfo, 2);
	radio_box->type = GNOME_APP_UI_RADIOITEMS;
	radio_box->label = _("Servers");
	radio_box->hint = _("Servers");
	radio_box->moreinfo = buf;
	radio_box[1] = end;


	/* update the main menu */
	if (server_menu)
		gnome_app_remove_menus (GNOME_APP(Pan.window), "Server/", server_menu_len);
	gnome_app_insert_menus (GNOME_APP(Pan.window), "Server/", radio_box);

	/* update the local varialbes */
	g_free (server_menu);
	g_free (server_menu_radiobox);
	server_menu = buf;
	server_menu_radiobox = radio_box;
	server_menu_len = i;

	/* save groups to disk */
	server_list_save ();

	/* cleanup */
	g_string_free (str, TRUE);
}

Server*
server_fetch (const gchar *name)
{
	Server *sdata = NULL;
	gboolean has_section;

	gchar* buf = g_strdup_printf ( "/Pan/%s/", name );
	gnome_config_push_prefix ( buf );
	has_section = gnome_config_has_section ( buf );
	g_free ( buf );

	if ( has_section )
	{
		sdata = g_new0 (Server, 1);
		sdata->address = gnome_config_get_string ("Address");
		if (sdata->address == NULL)
		{
			g_free (sdata);
			sdata = NULL;
		}
		else
		{
			gchar *path;
			
			/* make sure the server's config directory exists */
		       	path = g_strdup_printf ("/%s/%s/", data_dir, name);
			directory_check (path);
			g_free (path);

			/* load the sdata from gnome_config */
			sdata->port =
				gnome_config_get_int ("Port=119");
			sdata->name =
				g_strdup(name);
			sdata->gen_msgid =
				gnome_config_get_int ("Generate_Msgid=0");
			sdata->need_auth =
				gnome_config_get_int ("Auth_Required=0");
			sdata->username =
				gnome_config_get_string ("Username");
			sdata->password =
				gnome_config_get_string ("Password");
			sdata->max_connections =
				gnome_config_get_int ("Max_Connections=2");
			sdata->idle_secs_before_timeout =
				gnome_config_get_int ("Idle_Timeout_Secs=180");
			sdata->groups = g_ptr_array_new ();
		}
	}
	gnome_config_pop_prefix ();

	return sdata;
}

/**
 * server_free:
 * @sdata: server to be free'd
 *
 * Free all the memory used by a Server struct
 **/
void
server_free (Server *sdata)
{
	if (sdata != NULL)
        {
		g_free (sdata->name);
		g_free (sdata->address);
		g_free (sdata->username);
		g_free (sdata->password);
		g_free (sdata);
	}
}

void
server_init (void)
{
	gchar *last = NULL;
	GSList *l = NULL;
	int i;

	/* This gets called before gtk_main and
	 * at the end of guru_finish */
	Pan.serverlist = server_list_load ();
	server_menu_update ();

	/* try hard to select a server.. if there's one in the ini file, use
	 * it; otherwise, go with the first one in the list (if there is one)
	 */
	last = gnome_config_get_string ("/Pan/State/Server");
	if (!last && Pan.serverlist)
		last = g_strdup (((Server*)Pan.serverlist->data)->name);
	if (last!=NULL)
	{
		for (i=0, l=Pan.serverlist; l; l=l->next, ++i)
		{
			Server *sdata = SERVER(l->data);
			if (!strcmp(sdata->name, last))
			{
				gtk_check_menu_item_set_active (
		                	GTK_CHECK_MENU_ITEM(
						server_menu[i].widget), TRUE);
				grouplist_server_selected (sdata);
				break;
			}
		}
		g_free (last);
	}
}


static void
server_list_save (void)
{
	GString * buf = g_string_new (NULL);
	GString * servers = g_string_new (NULL);
	GSList *l = NULL;
	gchar * p;

	/* Out with the old */
	p = gnome_config_get_string ("/Pan/Servers/Names");
	if (p != NULL)
	{
		char **sd = g_strsplit (p, ",", -1);
		int i;
		for (i=0; sd[i]; i++)
		{
			g_string_sprintf (buf, "/Pan/%s", sd[i]);
			gnome_config_clean_section (buf->str);
		}
		g_strfreev (sd);
		g_free (p);
	}
	gnome_config_sync ();

	/* In with the new */
	for (l=Pan.serverlist; l!=NULL; l=l->next)
	{
		Server *s = SERVER(l->data);

		g_string_sprintf (buf, "/Pan/%s/Address", s->name);
		gnome_config_set_string (buf->str, s->address);

		g_string_sprintf (buf, "/Pan/%s/Port", s->name );
		gnome_config_set_int (buf->str, s->port);

		g_string_sprintf (buf, "/Pan/%s/Auth_Required", s->name);
		gnome_config_set_int (buf->str, s->need_auth);

		g_string_sprintf (buf, "/Pan/%s/Generate_Msgid", s->name);
		gnome_config_set_int (buf->str, s->gen_msgid);

		g_string_sprintf (buf, "/Pan/%s/Username", s->name);
		gnome_config_set_string (buf->str, s->username);

		g_string_sprintf (buf, "/Pan/%s/Password", s->name);
		gnome_config_set_string (buf->str, s->password);

		g_string_sprintf (buf, "/Pan/%s/Max_Connections", s->name);
		gnome_config_set_int (buf->str, s->max_connections);

		g_string_sprintf (buf, "/Pan/%s/Idle_Timeout_Secs", s->name);
		gnome_config_set_int (buf->str, s->idle_secs_before_timeout);

		if (servers->len)
			g_string_append_c (servers, ',');
		g_string_append (servers, s->name);
	}

	gnome_config_set_string ("/Pan/Servers/Names", servers->str);
	g_string_free (servers, TRUE);
	g_string_free (buf, TRUE);

	gnome_config_sync ();
}

static gint
server_name_compare (gconstpointer a, gconstpointer b)
{
	const Server* sa = (const Server*) a;
	const Server* sb = (const Server*) b;
	return g_strcasecmp (sa->name, sb->name);
}

/**
 * server_list_load:
 *
 * Load a list of servers from the user's ~/.gnome/Pan file
 *
 * Return value: GSList of Server's for each server found
 **/
GSList *
server_list_load (void)
{
	gchar *p;
	gchar **sd;
	gint i;
	GSList *l = NULL;

	if ((p = gnome_config_get_string ("/Pan/Servers/Names")) == NULL)
		return NULL;
    
	sd = g_strsplit (p, ",", 0);
	g_free (p);
	
	if (sd == NULL)
		return NULL;

	for (i = 0; sd[i]; i++) {
		Server *sdata = server_fetch (sd[i]);
		if (sdata)
			l = g_slist_insert_sorted (l, sdata, server_name_compare);
	}
	g_strfreev (sd);
	return l;
}




static int
compare_name_to_group (const void* a, const void* b)
{
	return strcmp ((const char*)a, (*(Group**)b)->name);
}

gboolean
server_has_group (const Server* server, const char* group_name)
{
	gboolean exact_match = FALSE;

	g_return_val_if_fail (server!=NULL, FALSE);
	g_return_val_if_fail (group_name!=NULL, FALSE);
	g_return_val_if_fail (*group_name!='\0', FALSE);

	lower_bound (group_name,
                     server->groups->pdata,
                     server->groups->len,
                     sizeof(gpointer),
                     compare_name_to_group,
                     &exact_match);

	return exact_match;
}

