/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "globals.h"
#include "acache.h"
#include "article.h"
#include "article-db.h"
#include "debug.h"
#include "log.h"
#include "message-window.h"
#include "nntp.h"
#include "queue.h"
#include "task-bodies.h"
#include "text.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static int task_bodies_run (Task* item);

static gchar* task_bodies_describe (const StatusItem* item);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

static void
task_bodies_destructor (PanObject *o)
{
	TaskBodies *item = TASK_BODIES(o);

	/* destroy TaskBodies parts */
	pan_g_ptr_array_foreach (item->articles, (GFunc)pan_object_unref, NULL);
	g_ptr_array_free (item->articles, TRUE);
	item->articles = NULL;
	item->working_index = 0;

	/* destroy parent class */
	task_destructor (o);
}

/************
*************  PUBLIC ROUTINES
************/

PanObject*
task_bodies_new (Server      * server,
                 Group       * group,
                 GPtrArray   * articles)
{
	TaskBodies* bodies = NULL;

	/* sanity checks */
	g_return_val_if_fail (server!=NULL, NULL);
	g_return_val_if_fail (group!=NULL, NULL);
	g_return_val_if_fail (articles!=NULL, NULL);

	/* create the object */
       	bodies = g_new0 (TaskBodies, 1);
        debug (DEBUG_PAN_OBJECT, "task_bodies_new: %p", bodies);

	/* initialize the parent class */
	task_constructor (TASK(bodies),
		task_bodies_destructor,
		task_bodies_describe,
		task_bodies_run,
		server, TRUE, TRUE);

	/* initialize the task-bodies */
	bodies->group = group;
	bodies->articles = articles;
	bodies->working_index = 0;
	pan_g_ptr_array_foreach (articles, (GFunc)pan_object_ref, NULL);

	return PAN_OBJECT(bodies);
}

/*****
******
*****/

static gchar*
task_bodies_describe (const StatusItem* si)
{
	TaskBodies *item = TASK_BODIES(si);

	return g_strdup_printf (_("Downloading %d articles from group %s"),
		item->articles->len,
		item->group->name );
}

/*****
******
*****/

static int
task_bodies_run (Task* item)
{
	Server * const server = item->sdata;
	Group * const group = TASK_BODIES(item)->group;
	const gboolean * abort = &item->hint_abort;
	GPtrArray * const articles = TASK_BODIES(item)->articles;
	int *index = &TASK_BODIES(item)->working_index;
	article_db db;
	int step = 0;


	/* sanity checks */
	g_return_val_if_fail(server!=NULL, -1);
	g_return_val_if_fail(group!=NULL, -1);
	g_return_val_if_fail(articles!=NULL, -1);


	/* go to the right group */
	if (*abort || nntp_set_group(item->sock, group))
	{
		status_item_emit_error (
			STATUS_ITEM(item),
			_("Couldn't select group '%s'"),
			group->name);

		return -1;
	}


	/* let the listeners know how far along we are */
	if (TRUE)
	{
		int i;
		int line_count_total = 0;
		int line_count_remaining = 0;
		for (i=0; i<articles->len; ++i) {
			const int lines = ARTICLE(g_ptr_array_index(articles,i))->linecount;
			line_count_total += lines;
			if (i>=*index)
				line_count_remaining += lines;
		}
		status_item_emit_init_steps (STATUS_ITEM(item), line_count_total);
		step = line_count_total - line_count_remaining;
		status_item_emit_set_step (STATUS_ITEM(item), step);
	}


	/* download the articles */
	for (; *index<articles->len && !*abort; ++(*index))
	{
		Article* article = ARTICLE(g_ptr_array_index(articles,*index));
		int retval = 0;

		if (!acache_body_exists(article->message_id) ||
		    !acache_header_exists(article->message_id))
		{
			retval = nntp_article_download (
				STATUS_ITEM(item),
				TASK(item)->sock,
				server, group, article,
				&TASK(item)->hint_abort,
				FALSE);
		}

		if (retval)
		{
			status_item_emit_status(STATUS_ITEM(item),
				_("Downloading article ``%s'' failed."),
				article->subject);
		}
		else
		{
			step += article->linecount;
			status_item_emit_set_step (STATUS_ITEM(item), step);

			status_item_emit_status (
				STATUS_ITEM(item),
				_("Getting article %d of %d"),
				1+*index,
				articles->len);

			article_remove_flag (
				server, group, article,
				STATE_DOWNLOAD_QUEUED, FALSE);
		}
	}

	/* sync the article database */
	db = ahdb_ref (server, group);
	ahdb_sync (db);
	ahdb_unref (db);

	return 0;
}
