/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "globals.h"
#include "acache.h"
#include "article.h"
#include "article-db.h"
#include "debug.h"
#include "log.h"
#include "message-window.h"
#include "nntp.h"
#include "queue.h"
#include "task-body.h"
#include "text.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static int task_body_run (Task* item);

static gchar* task_body_describe (const StatusItem* item);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

static void
task_body_destructor (PanObject *o)
{
	TaskBody *item = TASK_BODY(o);

	if (item->checkout) {
		item->checkout = FALSE;
		acache_file_checkin(item->article->message_id);
	}

	pan_object_unref (PAN_OBJECT(item->article));
	item->article = NULL;

	task_destructor (o);
}

/************
*************  PUBLIC ROUTINES
************/

PanObject*
task_body_new_defaults (Server   * sdata,
                        Group    * gdata,
                        Article  * article)
{
	return task_body_new (sdata, gdata, article, FALSE, NNTP_ARTICLE_READ);
}

PanObject*
task_body_new (Server             * sdata,
               Group              * gdata,
               Article            * article,
               gboolean             open_in_new_window,
               MessageWindowType    message_window_type)
{
	TaskBody* body = NULL;
	gboolean needs_socket;

	/* sanity checks */
	g_return_val_if_fail (article!=NULL, NULL);
	g_return_val_if_fail (article->message_id!=NULL, NULL);
	g_return_val_if_fail (*article->message_id!='\0', NULL);

	pan_object_ref (PAN_OBJECT(article));

	/* do we need to pull this body from the network? */
	needs_socket = TRUE;
	if (acache_body_exists(article->message_id) &&
	    acache_header_exists(article->message_id))
	{
        	debug (DEBUG_PAN_OBJECT, "task_body_new: no socket needed; it's cached");
		acache_file_checkout(article->message_id);
		needs_socket = FALSE;
	}

	/* create the object */
       	body = g_new0 (TaskBody, 1);
        debug (DEBUG_PAN_OBJECT, "task_body_new: %p", body);

	/* initialize the parent class */
	task_constructor (TASK(body),
		task_body_destructor,
		task_body_describe,
		task_body_run,
		sdata, TRUE, needs_socket);
	TASK(body)->is_article_reader = TRUE;


	/* initialize the task-body */
	debug (DEBUG_PAN_OBJECT, "task_body: %p", article);
	body->gdata = gdata;
	body->article = article;
	body->open_in_new_window = open_in_new_window;
	body->message_window_type = message_window_type;
	body->checkout = !needs_socket;

	return PAN_OBJECT(body);
}

/*****
******
*****/

static gchar*
task_body_describe (const StatusItem* si)
{
	TaskBody *item = TASK_BODY(si);

	if (item->checkout)
	{
		return g_strdup_printf (_("Loading body of cached article ``%s'' from group %s"),
			item->article->message_id,
			item->gdata->name );
	} else {
		return g_strdup_printf (_("Downloading body of ``%s'' from group %s"),
			item->article->message_id,
			item->gdata->name );
	}
}

/*****
******
*****/

static int
task_body_run (Task* item)
{
	Server * const sdata = item->sdata;
	Group * const gdata = TASK_BODY(item)->gdata;
	Article * const article = TASK_BODY(item)->article;
	const gboolean * const abort = &TASK(item)->hint_abort;

	/* sanity checks */
	g_return_val_if_fail(sdata!=NULL, -1);
	g_return_val_if_fail(gdata!=NULL, -1);
	g_return_val_if_fail(article!=NULL, -1);
	g_return_val_if_fail(article->message_id!=NULL, -1);
	g_return_val_if_fail(*article->message_id!='\0', -1);

	if (TASK_BODY(item)->checkout)
		status_item_emit_status(STATUS_ITEM(item),
			_("Loading Cached Article '%s'"),
			article->message_id );

	article_remove_flag (sdata, gdata, article, STATE_DOWNLOAD_QUEUED, TRUE);

	/* Check the cache before we download */
	if (!acache_header_exists(article->message_id) ||
            !acache_body_exists(article->message_id))
	{
		int val = 0;

		acache_file_checkout (article->message_id);
	        val = nntp_set_group(TASK(item)->sock, gdata);
	        if (*abort || val)
		{
			g_message ("abort: %d val: %d", *abort, val);
			status_item_emit_error (
				STATUS_ITEM(item),
				_("Couldn't select group '%s'"),
				gdata->name);

			return -1;
		}

		if (nntp_article_download(
			STATUS_ITEM(item), TASK(item)->sock,
			sdata, gdata, article, &TASK(item)->hint_abort, TRUE))
		{
			status_item_emit_status(STATUS_ITEM(item),
				_("Downloading article ``%s'' failed."),
				article->subject);
			acache_file_checkin (article->message_id);

			/* if user's expecting a message, let them know it's off. */
			if (item->high_priority && STATUS_ITEM(item)->last_error!=NULL)
				update_body_pane (Pan.text, STATUS_ITEM(item)->last_error);

			return -1;
		}
	}

	/**
	***  Display the message?
	**/

	if (TASK_BODY(item)->open_in_new_window /* message window */
		|| (item->high_priority && use_message_window_always))
	{
		Message* m = message_new (sdata, gdata, article);
		m->type = TASK_BODY(item)->message_window_type;
		message_window_new (m);
		article_mark_read (sdata, gdata, article, TRUE);
	}
	else if (item->high_priority) /* use default text window */
	{
		text_update (article);

		/* check in this file.. */
		if (TASK_BODY(item)->checkout) {
			TASK_BODY(item)->checkout = FALSE;
			acache_file_checkin (article->message_id);
		}
		article_mark_read (sdata, gdata, article, TRUE);
	}

	/* cleanup */
	acache_file_checkin (article->message_id);

	return 0;
}
