/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/libgnomeui.h>

#include "acache.h"
#include "article.h"
#include "article-db.h"
#include "debug.h"
#include "decode.h"
#include "nntp.h"
#include "prefs.h" /* for download_dir */
#include "queue.h"
#include "task-decode.h"
#include "util.h" /* for pan_lock/unlock */

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint task_decode_run (Task* item);

static char* task_decode_describe (const StatusItem* item);

static int nntp_article_decode (Task *item);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

static char*
task_decode_describe (const StatusItem* item)
{
	TaskDecode *decode = TASK_DECODE(item);

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

        return decode->open
		? g_strdup_printf (_("Opening %s"), decode->readable_name)
		: g_strdup_printf (_("Saving %s"), decode->readable_name);
}

/************
*************  PUBLIC ROUTINES
************/

/*****
******
*****/

static void
task_decode_destructor (PanObject* object)
{
	TaskDecode* item = TASK_DECODE(object);

        debug (DEBUG_PAN_OBJECT, "task_decode destructor: %p", object);

	/* destruct the TaskDecode parts */
	g_free (item->readable_name);
	item->readable_name = NULL;
	g_free (item->path);
	item->path = NULL;
	g_free (item->filename);
	item->filename = NULL;
	g_slist_foreach (item->articles, (GFunc)pan_object_unref, NULL);
	g_slist_free (item->articles);
	item->articles = NULL;

	/* destruct the superclass */
	task_destructor (PAN_OBJECT(object));
}

PanObject*
task_decode_new (Server* sdata,
                 Group* gdata,
                 const char* readable_name, /* for description */
                 const char* path,
                 const char* filename,
                 const GSList* articles,
                 gboolean open)
{
	TaskDecode *item = g_new0 (TaskDecode, 1);

        debug (DEBUG_PAN_OBJECT, "task_decode_new: %p", item);

	task_constructor (TASK(item),
		task_decode_destructor,
		task_decode_describe,
		task_decode_run,
		sdata, FALSE, TRUE);

	TASK(item)->is_leech = TRUE;

	/* construct this class... */
	item->gdata = gdata;
	item->open = open;
	item->readable_name = g_strdup(readable_name);
	item->path = g_strdup (path);
	item->filename = g_strdup (filename);
	item->articles = g_slist_copy ((GSList*)articles);
	g_slist_foreach (item->articles, (GFunc)pan_object_ref, NULL);

	/* if no path specified, fall back on Pan group and global defaults */
	if (!item->path)
	       	item->path = group_get_attrib_string (
			sdata, gdata, "download_dir");
	if (!item->path)
		item->path = g_strdup(download_dir);

	return PAN_OBJECT(item);
}

/*****
******
*****/


/*****
******
*****/

static gint
task_decode_run (Task* item)
{
	GSList *l;
	Server *sdata = item->sdata;
	Group *gdata = TASK_DECODE(item)->gdata;
	GSList *articles = TASK_DECODE(item)->articles;
	decode_data *dd = NULL;
	Article* adata;

	/* Checkout the article bodies so they don't get
	 * expired while we're using them */
	for (l=articles; l!=NULL; l=l->next)
		acache_file_checkout (ARTICLE(l->data)->message_id);

	/* get the articles bodies */
	if (nntp_article_decode(item))
		return -1;

	/* update "read" attribute */
       	adata = ARTICLE(articles->data);
	if (!article_flag_on (adata, STATE_READ))
		article_mark_read (sdata, gdata, adata, TRUE);

	/* pass it off to article_decode */
	dd = g_malloc (sizeof(decode_data));
	dd->sdata = sdata;
	dd->gdata = gdata;
	dd->articles = articles;
	dd->open = TASK_DECODE(item)->open;
	dd->item = STATUS_ITEM(item);
	dd->path = TASK_DECODE(item)->path;
	dd->filename = TASK_DECODE(item)->filename;
	decode_article (dd);
	g_free (dd);

	/* Check the bodies back in -- make them
	 * eligible for expire again */
	for (l=articles; l!=NULL; l=l->next)
		acache_file_checkin (ARTICLE(l->data)->message_id);

	return 0;
}
/*****
******
*****/

static int
nntp_article_decode (Task *item)
{
	PanSocket *sock = item->sock;
	const gboolean * const abort = &TASK(item)->hint_abort;
	Server *sdata = item->sdata;
	Group *gdata = TASK_DECODE(item)->gdata;
	GSList *articles = TASK_DECODE(item)->articles;
	const char *buffer = NULL;
	float count = 0;
	GSList *l = NULL;
	gint total_line_count = 0;
	gint lines_so_far = 0;

	/* sanity check */
	g_return_val_if_fail (sock!=NULL, 0);
	g_return_val_if_fail (sdata!=NULL, 0);
	g_return_val_if_fail (gdata!=NULL, 0);
	g_return_val_if_fail (articles!=NULL, 0);

	/* set the group */
	if (*abort || nntp_set_group (sock, gdata)) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't select group '%s'"), gdata->name );
		return -1;
	} 

	/* get the total line count */
	total_line_count = 0;
	for (l=articles; l!=NULL; l=l->next)
		total_line_count += ARTICLE(l->data)->linecount;

	/* Download bodies for all of the parts */
	lines_so_far = 0;
	for ( l=articles; l!=NULL; l=l->next)
	{
		FILE *f = NULL;
		Article* adata0 = ARTICLE(l->data);

		/* see if we already have this article/part cached */
		if (acache_body_exists (adata0->message_id))
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("already have %s"),
				adata0->subject);

			lines_so_far += adata0->linecount;

			status_item_emit_progress (STATUS_ITEM(item),
				(100.0*lines_so_far)/total_line_count);

			continue;
		}

		/* let listeners know what we're doing */
		if (adata0->parts > 1)
			status_item_emit_status (STATUS_ITEM(item),
				_("Downloading %d of %d: %s"),
				adata0->part, adata0->parts, adata0->subject);
		else
			status_item_emit_status (STATUS_ITEM(item),
				_("Downloading: %s"), adata0->subject);

		/* download the article */
		if (*abort || pan_socket_putline_va(sock, "BODY %s\r\n",adata0->number))
		{
			task_emit_sockwrite_err(TASK(item));
			return -1;
		}
		if (*abort || pan_socket_getline (sock, &buffer))
		{
			task_emit_sockwrite_err(TASK(item));
			return -1;
		}
		if (strtol(buffer, (char **) NULL, 10) != 222) {
			article_add_flag (
				sdata, gdata, adata0, STATE_ERROR, TRUE);
			return -1;
		}

		count = 1;
		if (*abort || pan_socket_getline (sock, &buffer))
		{
			task_emit_sockread_err(TASK(item));
			return -1;
		}
	
		f = acache_open_body (adata0);
		while (strncmp(buffer, ".\r\n", 3))
		{
			acache_putline_body (f, buffer);
			if (*abort || pan_socket_getline (sock, &buffer))
			{
				task_emit_sockread_err(TASK(item));
				acache_close_body (f);
				acache_delete (adata0->message_id);
				return -1;
			}

			/* emit progress every once in awhile */
			if (++lines_so_far % 47 == 0)
				status_item_emit_progress (STATUS_ITEM(item),
					(100.0*lines_so_far)/total_line_count);
		}
		acache_close_body (f);
	}

	/* no errors received; mark as read */
	if (articles != NULL)
		article_mark_read (sdata, gdata, ARTICLE(articles->data), TRUE);

	return 0;
}


/*****
******
******  VALIDATE-AND-QUEUE or SELF-DESTRUCT
******
*****/

typedef struct
{
	GtkWidget* clist;
	GtkWidget* window;
	TaskDecode* item;
}
ResolvePartsUI;

static void
select_articles_clicked_cb (GnomeDialog* dialog,
                            int button,
                            gpointer user_data )
{
	ResolvePartsUI *ui = (ResolvePartsUI*) user_data;

	if (button==0) /* "ok" */
	{
		GList* sel = GTK_CLIST(ui->clist)->selection;
		const int rows = GTK_CLIST(ui->clist)->rows;
		int row;

		/* remove the non-selected articles */
		for (row=rows-1; row>=0; --row)
		{
			gboolean row_selected = g_list_find (sel,GINT_TO_POINTER(row))!=NULL;

			if (!row_selected)
			{
				GSList* n = g_slist_nth (ui->item->articles, row);
				g_message ("not using adata [%s]", ARTICLE(n->data)->subject);

				ui->item->articles = g_slist_remove_link(ui->item->articles,n);

				g_slist_free_1 (n);
				pan_object_unref (PAN_OBJECT(n->data));
			}
		}

		queue_add (TASK(ui->item));
	}
	else
	{
		pan_object_unref (PAN_OBJECT(ui->item));
	}

	/* cleanup */
	pan_lock ();
	gtk_widget_destroy (ui->window);
	pan_unlock ();
	g_free (ui);
}

static void
choose_article_parts_auto (TaskDecode* item)
{
	GSList* l;
	Article *last_adata = NULL;
	
	l = item->articles;
	while (l != NULL)
	{
		Article *adata = ARTICLE (l->data);
		
		l = l->next;
		
		if (last_adata && (last_adata->part == adata->part))
		{
			/* Select the better of the two assuming
			 * that more lines is better */
			if (last_adata->linecount > adata->linecount) {
				item->articles = g_slist_remove (item->articles, adata);
				pan_object_unref (PAN_OBJECT(adata));
			} else if (last_adata->linecount < adata->linecount) {
				item->articles = g_slist_remove (item->articles, last_adata);
				pan_object_unref (PAN_OBJECT(last_adata));
				last_adata = adata;
			} else {
				/* Check the date next if lines match,
				 * last posted is preferred */
				if (last_adata->date > adata->date) {
					item->articles = g_slist_remove (item->articles, adata);
					pan_object_unref (PAN_OBJECT(adata));
				} else { 
					/* if adata is newer, or the same, kill the other one */
					item->articles = g_slist_remove (item->articles, last_adata);
					pan_object_unref (PAN_OBJECT(last_adata));
					last_adata = adata;
				}	
			}
		} else {
			last_adata = adata;
		}
	}

	queue_add (TASK(item));
}

static void
choose_article_parts_manual (TaskDecode* item)
{
	GSList* l;
	GtkWidget* w;
	GtkWidget* list;
	GtkWidget* vbox;
	GtkWidget* label;
	char* columns [4];
	int last_part = -1;
	ResolvePartsUI* ui = g_new0 (ResolvePartsUI, 1);

	ui->item = item;

	pan_lock ();

	ui->window = gnome_dialog_new (
		_("Decode Multipart Message"),
		GNOME_STOCK_BUTTON_OK,
		GNOME_STOCK_BUTTON_CANCEL,
		NULL);
	gtk_signal_connect (
		GTK_OBJECT(ui->window), "clicked",
		GTK_SIGNAL_FUNC(select_articles_clicked_cb), ui);

	gtk_window_set_policy (GTK_WINDOW (ui->window), TRUE, TRUE, TRUE);
	gnome_dialog_set_default (GNOME_DIALOG(ui->window), 1);
	vbox = GNOME_DIALOG(ui->window)->vbox;

	label = gtk_label_new (
		_("Pan found multiple copies of some of the parts of this post.\n"
		  "Please select the parts to use for this decode."));
	gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, GNOME_PAD_SMALL);

	columns[0] = _("Subject");
	columns[1] = _("Lines");
	columns[2] = _("Author");
	columns[3] = _("Date");
	ui->clist = list = gtk_clist_new_with_titles (4, columns);
	gtk_clist_set_selection_mode (GTK_CLIST(list), GTK_SELECTION_MULTIPLE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(list), 0, TRUE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(list), 1, TRUE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(list), 2, TRUE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(list), 3, TRUE);

	for (l=item->articles; l!=NULL; l=l->next)
	{
		char datebuf[32];
		char lines[8];
		Article* adata = ARTICLE (l->data);
		const struct tm local_tm = *localtime (&adata->date);
		char* text[4] = { adata->subject, lines, adata->author, datebuf };
		int row;
		sprintf (lines, "%d", adata->linecount);
		strftime (datebuf, sizeof(datebuf), "%Y/%m/%d %H:%M", &local_tm);
		row = gtk_clist_append (GTK_CLIST(list), text);
		gtk_clist_set_row_data (GTK_CLIST(list), row, adata);
		if (last_part != adata->part)
		{
			gtk_clist_select_row (GTK_CLIST(list), row, 0);
			last_part = adata->part;
		}
	}
	w = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(w), ui->clist);
	gtk_box_pack_start (GTK_BOX(vbox), w, TRUE, TRUE, GNOME_PAD_SMALL);

	gtk_window_set_default_size (GTK_WINDOW (ui->window), 500, 400);
	gtk_widget_show_all (ui->window);

	pan_unlock ();
}

void
task_decode_validate_and_queue_or_self_destruct (TaskDecode* item)
{
	const int length = g_slist_length(item->articles);

	if ((length > 1) && (length != ARTICLE(item->articles->data)->parts))
	{
		if (auto_part_selection)
		{
			choose_article_parts_auto (item);	
		}
		else
		{
			choose_article_parts_manual (item);
		}
	}
	else
	{
		queue_add (TASK(item));
	}
}

