/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>

#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-config.h>

#include "article.h"
#include "article-db.h"
#include "articlelist.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "grouplist.h"
#include "task-grouplist.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint task_grouplist_run (Task* item);

static int nntp_grouplist_download_all (TaskGrouplist* item, GPtrArray* setme);

static int nntp_grouplist_download_new (TaskGrouplist* item, GPtrArray* setme);

static gchar* task_grouplist_describe (const StatusItem* item);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

PanObject*
task_grouplist_new (Server *sdata,
                    gboolean high_priority,
                    GrouplistDownloadType dl_type)
{
	TaskGrouplist *item = g_new0 (TaskGrouplist, 1);

	debug (DEBUG_PAN_OBJECT, "task_grouplist_new: %p", item);
	
	/* construct superclass... */
	task_constructor (TASK(item),
		task_destructor,
		task_grouplist_describe,
		task_grouplist_run, sdata, high_priority, TRUE);

	/* construct this class... */
	item->sdata = sdata;
	item->download_type = dl_type;

	return PAN_OBJECT(item);
}

/*****
******
*****/

static gchar*
task_grouplist_describe (const StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

	switch (TASK_GROUPLIST(item)->download_type) {
		case GROUPLIST_ALL: action=_("all groups"); break;
		case GROUPLIST_NEW: action=_("new groups"); break;
		default: pan_warn_if_reached(); action=_("BUG IN CODE"); break;
	}

	return g_strdup_printf ( _("Getting %s from %s"),
		action,
		TASK_GROUPLIST(item)->sdata->name );
}

static void
update_last_grouplist_update_time (const Server* sdata)
{
	gchar *tmp;
	struct tm *gmt;
	gchar date[32]; /* "YYYYMMDD HHMMSS GMT" */
	time_t secs;

	time (&secs);
	gmt = gmtime (&secs);
	strftime (date, sizeof (date), "%Y%m%d %H%M%S GMT", gmt);

	tmp = g_strdup_printf ("/Pan/%s/Last_Grouplist_Update", sdata->name);
	gnome_config_set_string (tmp, date);
	g_free (tmp);

	gnome_config_sync ();
}

static int
compare_pgroup_ppgroup_name (const void* a, const void* b)
{
	return strcmp (((Group*)a)->name, (*(Group**)b)->name);
}

static gint
task_grouplist_run (Task *item)
{
	TaskGrouplist *item_g = TASK_GROUPLIST(item);
	Server *server = item->sdata;
	GPtrArray *groups = g_ptr_array_new ();
	int retval;

	status_item_emit_status (STATUS_ITEM(item),
		_("Retrieved 0 groups for '%s'"), server->name);

	if (item_g->download_type==GROUPLIST_ALL)
	{
		retval = nntp_grouplist_download_all (item_g, groups);
	}
	else if (item_g->download_type==GROUPLIST_NEW)
	{
		retval = nntp_grouplist_download_new (item_g, groups);
	}
	else
	{
		pan_warn_if_reached ();
		return -1;
	}

	/* update the list if we had a success. */
	if (!retval)
	{
		int i;
		gboolean changed = FALSE;

		/* for the next time we call GROUPLIST_NEW... */
		update_last_grouplist_update_time (server);

		status_item_emit_status (STATUS_ITEM(item),
			_("Merging groups for `%s'"), server->name);

		/* add the groups we don't already have */
		for (i=0; i<groups->len; ++i)
		{
			Group* group = (Group*) g_ptr_array_index (groups, i);
			gboolean exact_match = FALSE;
			int index = lower_bound (group,
                                                 server->groups->pdata,
                                                 server->groups->len,
                                                 sizeof(gpointer),
                                                 compare_pgroup_ppgroup_name,
                                                 &exact_match);

			if (exact_match) /* already have group */
			{
				group_free (group);
			}
			else
			{
				changed = TRUE;
				group->flags |= GROUP_NEW;
				pan_g_ptr_array_insert (server->groups, group, index);
			}
		}

		if (changed)
		{
			/* server group has changed; update the db & ui */
			grouplist_save (server);
			grouplist_update_ui (server);
		}
		else
		{
			g_message (_("No change in groups."));
		}
	}

	/* cleanup */
	g_ptr_array_free (groups, TRUE);
	return retval;
}


/*---[ nntp_grouplist_download ]--------------------------------------
 *
 *--------------------------------------------------------------------*/
static int
nntp_grouplist_download_all (TaskGrouplist *item, GPtrArray *groups)
{
	const gboolean * const abort = &TASK(item)->hint_abort;
	PanSocket *sock = TASK(item)->sock;
	const Server *sdata = TASK(item)->sdata;
	gboolean get_descriptions = TRUE;
	const char *buffer = NULL;
	int read_status = 0;
	int description_count = 0;
	
	status_item_emit_status (STATUS_ITEM(item),
		_("Retrieved 0 groups for '%s'"), sdata->name);

	if (*abort || pan_socket_putline (sock, "LIST\r\n")) {
		task_emit_sockwrite_err (TASK(item));
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		task_emit_sockread_err (TASK(item));
		return -1;
	}
	if (atoi(buffer)!=215) { /* 215 = list follows */
		status_item_emit_error (STATUS_ITEM(item),
			_("Group list could not be retrieved\nServer said: %s"),
			buffer);
		return -1;
	}

	while (!*abort && (((read_status = pan_socket_getline (sock, &buffer))) == 0))
	{
		const gchar *pch;
		Group *group = NULL;
	
		if (!strcmp(buffer,".\r\n"))
			break;
		
		/* init the group */
		group = group_new ();

		/* get the name */
		pch = strchr (buffer, ' ');
		if (!pch) pch = buffer + strlen(buffer);
		group->name = g_memdup (buffer, (pch-buffer)+1);
		group->name[pch-buffer] = '\0';

		/* update the list */
		g_ptr_array_add (groups, group);

		/* periodic status feedback */
		if (groups->len % 43 == 0)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Retrieved %d groups for '%s'"),
				groups->len, sdata->name);
		}
	}
	if (read_status) {
		/*FIXME: prompt whether or not to keep the list.  For now, leak. */
		task_emit_sockread_err (TASK(item));
		return -1;
	}

	/* sort so that we can bsearch them... */
	qsort (groups->pdata, groups->len,
	       sizeof(Group*),
	       compare_Group_std);

	/* now try to get the descriptions...
	   this is implemented as a second pass because the overlap between
	   LIST and LIST NEWSGROUPS isn't well defined; the latter may only
	   show the groups with a description, moreover, it may not be
	   available on older servers.  To make sure we get all the groups,
	   and descriptions for as many as we can, we use both commands.
	   -- csk */

	if ((read_status=pan_socket_putline (sock, "LIST NEWSGROUPS *\r\n")))
	{
		task_emit_sockwrite_err (TASK(item));
		get_descriptions = FALSE;
	}
	if (*abort || (get_descriptions && (read_status=pan_socket_getline (sock, &buffer))))
	{
		task_emit_sockread_err (TASK(item));
		get_descriptions = FALSE;
	}
	if (get_descriptions && atoi(buffer)!=215) /* 215 = list follows */
	{
		status_item_emit_error (STATUS_ITEM(item),
			_("Group list could not be retrieved\nServer said: %s"),
			buffer);
			get_descriptions = FALSE;
	}

	description_count = 0;

	while (get_descriptions && !read_status)
	{
		const gchar* pch = NULL;
		gchar *name = NULL;
		gchar *description = NULL;

		read_status = *abort ? 1 : pan_socket_getline (sock, &buffer);
		pch = buffer;

		if (read_status)
			continue;

		if (!strcmp(buffer,".\r\n"))
			break;	
		
		while (*pch && !isspace((int)*pch)) ++pch;
		name = g_strndup (buffer, pch-buffer);
		buffer = pch;
		if (( pch = strstr (buffer, "\r\n"))) {
			description = g_strndup (buffer, pch-buffer);
			g_strchug(g_strchomp (description));
		}

		++description_count;

		if (description_count % 43 == 0)
		{
			status_item_emit_status (
				STATUS_ITEM(item),
				_("Retrieved %d group descriptions for '%s'"),
				description_count, sdata->name);
		}

		if (name && description) {
			Group tmp;
			Group* ptmp = &tmp;
			int index;
			gboolean exact_match = FALSE;
			tmp.name = name;
			index = lower_bound (
				&ptmp,
				groups->pdata, groups->len,
				sizeof(Group*),
				compare_Group_std,
				&exact_match);
			if (exact_match)
				((Group*)groups->pdata[index])->description = g_strdup (description);
		}

		g_free (name);
		g_free (description);
	}
	if (read_status)
	{
		/*FIXME: prompt whether or not to keep the list.  For now, leak. */
		task_emit_sockread_err (TASK(item));
		return -1;
	}

	log_add_va (_("Retrieved %d groups"), groups->len);

	return 0;
}


/*--------------------------------------------------------------------
 * get "new" groups
 * 
 * based on RFC977 - JEL
 *--------------------------------------------------------------------*/
static int
nntp_grouplist_download_new (TaskGrouplist *item, GPtrArray *groups)
{
	PanSocket *sock = TASK(item)->sock;
	const gboolean * const abort = &TASK(item)->hint_abort;
	const Server *sdata = TASK(item)->sdata;
	const char *buffer = NULL;
	Group *gdata = NULL;
	int count = 0;
	gchar *buf;
	char datestr[64];
	gchar *pch;

	status_item_emit_status (STATUS_ITEM(item),
		_("Retrieving 0 new groups for '%s'"), sdata->name);

	/* get the date of the last grouplist update */
	buf = g_strdup_printf (
		"/Pan/%s/Last_Grouplist_Update=19700101 000000 GMT",
		sdata->name);
	pch = gnome_config_get_string_with_default (buf,  NULL);
	strcpy (datestr, pch);
	g_free (buf);
	g_free (pch);

	/* ask the server which groups have been added since then */
	if (pan_socket_putline_va (sock, "NEWGROUPS %s\r\n", datestr)) {
		task_emit_sockwrite_err (TASK(item));
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		task_emit_sockread_err (TASK(item));
		return -1;
	}

	if (atoi(buffer)==501) /* 501 Usage: NEWGROUPS yymmdd hhmmss ["GMT"] */
	{
		/* Paul McGarrry <mcgarray@tig.com.au> and his Newsmaster
		 * Simon Lyall <simon.lyall@ihug.co.nz> report that nntpcache
		 * isn't handling four-digit years.  So if the four-digit year
		 * fails, let's fall back to a two-digit year.
		 */
		if (pan_socket_putline_va(sock,"NEWGROUPS %s\r\n",datestr+2)) {
			task_emit_sockwrite_err (TASK(item));
			return -1;
		}
		if (*abort || pan_socket_getline (sock, &buffer)) {
			task_emit_sockread_err (TASK(item));
			return -1;
		}
	}
	if (atoi (buffer) != 231) /* 231 = list of new groups follows */
	{
		status_item_emit_error (STATUS_ITEM(item),
			_("Group list could not be retrieved\nServer said: %s"),
			buffer);
		return -1;
	}

	if (*abort || pan_socket_getline (sock, &buffer))
	{
		task_emit_sockread_err (TASK(item));
		return -1;
	}
	while (strncmp(buffer, ".\r\n", 3))
	{
		const gchar *pch;

		/* create a Group */
		gdata = group_new ();

		/* get the name */
		pch = strchr (buffer, ' ');
		if (!pch) pch = buffer + strlen(buffer);
		gdata->name = g_memdup (buffer, (pch-buffer)+1);
		gdata->name[pch-buffer] = '\0';
		gdata->flags |= GROUP_NEW;

		/* add the group to the list */
		g_ptr_array_add (groups, gdata);

		if (++count % 43 == 0)
		{
			status_item_emit_progress (STATUS_ITEM(item), count);

			status_item_emit_status (STATUS_ITEM(item),
				_("Retrieving %d new groups for '%s'"), 
				count, sdata->name);
		}

		if (*abort || pan_socket_getline (sock, &buffer))
		{
			task_emit_sockread_err (TASK(item));
			groups_free (groups);
			return -1;
		}
	}

	log_add_va (_("Retrieved %d new groups"), count);

	status_item_emit_status (STATUS_ITEM(item),
		_("Retrieved %d news groups for '%s'"), count, sdata->name);

	return 0;
}
